#!/usr/bin/env python
# Copyright 2020-2023 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Xing Zhang <zhangxing.nju@gmail.com>
#

import numpy as np
from pyscf import __config__
from pyscf import lib
from pyscf.lib import logger
from pyscf.pbc.symm import geom
from functools import reduce
from pyscf.pbc.symm.group import PGElement

SYMPREC = getattr(__config__, 'pbc_symm_space_group_symprec', 1e-6)
XYZ = np.eye(3)

def transform_rot(op, a, b, allow_non_integer=False):
    r'''
    Transform rotation operator from :math:`\mathbf{a}` basis system to :math:`\mathbf{b}` basis system.

    Note:
        This function raises error when the point-group symmetries of the two basis systems are different.

    Arguments:
        op : (3,3) array
            Rotation operator in :math:`\mathbf{a}` basis system.
        a : (3,3) array
            Basis vectors of :math:`\mathbf{a}` basis system (row-major).
        b : (3,3) array
            Basis vectors of :math:`\mathbf{b}` basis system (row-major).
        allow_non_integer : bool
            Whether to allow non-integer rotation matrix in the new basis system.
            Defualt value is False.

    Returns:
        A (3,3) array
            Rotation operator in :math:`\mathbf{b}` basis system.
    '''
    P = np.dot(np.linalg.inv(b.T), a.T)
    R = reduce(np.dot,(P, op, np.linalg.inv(P))).round(15)
    R[np.where(abs(R) < 1e-9)] = 0
    if not allow_non_integer:
        if(np.amax(np.absolute(R-R.round())) > SYMPREC):
            raise RuntimeError("Point-group symmetries of the two coordinate systems are different.")
        return R.round().astype(int)
    else:
        return R

def transform_trans(op, a, b):
    r'''
    Transform translation operator from :math:`\mathbf{a}` basis system to :math:`\mathbf{b}` basis system.

    Arguments:
        op : (3,) array
            Translation operator in :math:`\mathbf{a}` basis system.
        a : (3,3) array
            Basis vectors of :math:`\mathbf{a}` basis system (row-major).
        b : (3,3) array
            Basis vectors of :math:`\mathbf{b}` basis system (row-major).

    Returns:
        A (3,) array
            Translation operator in :math:`\mathbf{b}` basis system.
    '''
    P = np.dot(np.linalg.inv(b.T), a.T)
    return np.dot(op, P.T)


class SPGElement():
    '''
    Matrix representation of space group operations

    Attributes:
        rot : (d,d) array
            Rotation operator.
        trans : (d,) array
            Translation operator.
        dimension : int
            Dimension of the space: `d`.
    '''
    def __init__(self,
                 rot=np.eye(3, dtype=np.int32),
                 trans=np.zeros((3)), dimension=3):
        self.rot = np.asarray(rot)
        self.trans = np.asarray(trans)
        self.dimension = dimension
        if dimension != 3:
            raise NotImplementedError

    def dot(self, r_or_op):
        '''
        Operates on a point or multiplication of two operators
        '''
        if isinstance(r_or_op, np.ndarray) and r_or_op.ndim==1 and len(r_or_op)==3:
            return np.dot(r_or_op, self.rot.T) + self.trans
        elif isinstance(r_or_op, SPGElement):
            beta = self.rot
            b = self.trans
            alpha = r_or_op.rot
            a = r_or_op.trans
            op = SPGElement(np.dot(beta, alpha), b + np.dot(a, beta.T))
            return op
        else:
            raise KeyError("Input has wrong type: %s" % type(r_or_op))

    def dot_rot(self, r):
        '''
        Rotate a point (without translation)
        '''
        return np.dot(r, self.rot.T)

    def inv(self):
        '''
        Inverse of self
        '''
        inv_rot = np.linalg.inv(self.rot)
        trans = -np.dot(self.trans, inv_rot.T)
        return SPGElement(inv_rot, trans)

    def transform(self, a, b, allow_non_integer=False):
        r'''
        Transform from :math:`\mathbf{a}` basis system to :math:`\mathbf{b}` basis system.
        '''
        rot = transform_rot(self.rot, a, b, allow_non_integer)
        trans = transform_trans(self.trans, a, b)
        return SPGElement(rot, trans)

    @property
    def rot_is_eye(self):
        return ((self.rot - np.eye(self.dimension,dtype=int)) == 0).all()

    @property
    def rot_is_inversion(self):
        return ((self.rot + np.eye(self.dimension,dtype=int)) == 0).all()

    @property
    def trans_is_zero(self):
        return (np.abs(self.trans) < SYMPREC).all()

    @property
    def is_eye(self):
        '''
        Whether self is identity operation.
        '''
        return self.rot_is_eye and self.trans_is_zero

    @property
    def is_inversion(self):
        '''
        Whether self is inversion operation.
        '''
        return self.rot_is_inversion and self.trans_is_zero

    def __eq__(self, other):
        if not isinstance(other, SPGElement):
            raise TypeError
        return self.__hash__() == other.__hash__()

    def __ne__(self, other):
        if not isinstance(other, SPGElement):
            raise TypeError
        return not self.__eq__(other)

    def __lt__(self, other):
        if not isinstance(other, SPGElement):
            raise TypeError
        return self.__hash__() < other.__hash__()

    def __le__(self, other):
        if not isinstance(other, SPGElement):
            raise TypeError
        return self.__eq__(other) or self.__lt__(other)

    def __gt__(self, other):
        if not isinstance(other, SPGElement):
            raise TypeError
        return not self.__le__(other)

    def __ge__(self, other):
        if not isinstance(other, SPGElement):
            raise TypeError
        return not self.__lt__(other)

    def __hash__(self):
        r = PGElement.__hash__(self)

        trans = self.trans
        t = 0
        for i in range(self.dimension):
            t += int(np.round(trans[i] * 12.)) * 12**(2-i)

        return int(t * (3**(self.dimension ** 2)) + r)

    def a2b(self, cell):
        '''
        Transform from direct lattice system to reciprocal lattice system.
        '''
        return self.transform(cell.lattice_vectors(), cell.reciprocal_vectors())

    def a2r(self, cell):
        '''
        Transform from direct lattice system to Cartesian coordinate system.
        '''
        return self.transform(cell.lattice_vectors(), XYZ, True)

    def b2a(self, cell):
        '''
        Transform from reciprocal lattice system to direct lattice system.
        '''
        return self.transform(cell.reciprocal_vectors(), cell.lattice_vectors())

    def b2r(self, cell):
        '''
        Transform from reciprocal lattice system to Cartesian coordinate system.
        '''
        return self.transform(cell.reciprocal_vectors(), XYZ)

    def r2a(self, cell):
        '''
        Transform from Cartesian coordinate system to direct lattice system.
        '''
        return self.transform(XYZ, cell.lattice_vectors())

    def r2b(self, cell):
        '''
        Transform from Cartesian coordinate system to reciprocal lattice system.
        '''
        return self.transform(XYZ, cell.reciprocal_vectors())

    def __str__(self):
        s = ''
        for x in range(3):
            s += '%2d %2d %2d %10.6f\n' % (self.rot[x][0], self.rot[x][1], self.rot[x][2], self.trans[x])
        return s


class SpaceGroup(lib.StreamObject):
    '''
    Determines the space group of a lattice.

    Attributes:
        cell : :class:`Cell` object

        symprec : float
            Numerical tolerance for determining the space group.
            Default value is 1e-6 in the unit of length.
        verbose : int
            Print level. Default value equals to `cell.verbose`.
        backend: str
            Choose which backend to use for symmetry detection.
            Default is `pyscf` and other choices are `spglib`.
        ops : list of :class:`SPGElement` objects
            Matrix representation of the space group operations (in direct lattice system).
        nop : int
            Order of the space group.
        groupname : dict
            Standard symbols for symmetry groups.
            groupname['point_group_symbol']: point group symbol
            groupname['international_symbol']: space group symbol
            groupname['international_number']: space group number
    '''
    def __init__(self, cell, symprec=SYMPREC):
        self.cell = cell
        self.symprec = symprec
        self.verbose = cell.verbose
        self.stdout = cell.stdout
        self.backend = 'pyscf'

        # Followings are not input variables
        self.ops = []
        self.nop = 0
        self.groupname = {}

    def build(self, dump_info=True):
        if self.cell.dimension < 3 and self.backend == 'spglib':
            logger.warn(self, 'spglib only works for 3D system; '
                        'setting symmetry detection backend to pyscf native implementation.')
            self.backend = 'pyscf'

        if self.backend == 'spglib':
            from pyscf.pbc.symm.pyscf_spglib import cell_to_spgcell, get_symmetry_dataset, get_symmetry
            spgcell = cell_to_spgcell(self.cell)
            dataset = get_symmetry_dataset(spgcell, symprec=self.symprec)
            spg_symbol = dataset['international']
            spg_no = dataset['number']
            pg_symbol = dataset['pointgroup']
            symmetry = get_symmetry(spgcell, symprec=self.symprec)
            for rot, trans in zip(symmetry['rotations'], symmetry['translations']):
                self.ops.append(SPGElement(rot, trans))
        elif self.backend == 'pyscf':
            self.ops = geom.search_space_group_ops(self.cell, tol=self.symprec)
            pg_symbol = geom.get_crystal_class(self.cell, tol=self.symprec)[0]
            #TODO add space group symbol
            spg_symbol = None
            spg_no = None

        self.ops.sort()
        self.nop = len(self.ops)
        self.groupname['point_group_symbol'] = pg_symbol
        self.groupname['international_symbol'] = spg_symbol
        self.groupname['international_number'] = spg_no

        if dump_info:
            self.dump_info()
        return self

    def dump_info(self, ops=None):
        if ops is None:
            ops = self.ops
        if self.verbose >= logger.INFO:
            gn = self.groupname
            if gn['international_symbol'] is not None:
                logger.info(self, '[Cell] International symbol:  %s (%d)',
                            gn['international_symbol'], gn['international_number'])
            logger.info(self, '[Cell] Point group symbol:  %s', gn['point_group_symbol'])
        if self.verbose >= logger.DEBUG:
            if len(ops) < len(self.ops):
                message = 'Subset of space group symmetry operations:'
            else:
                message = 'Space group symmetry operations:'
            logger.debug(self, message)
            for op in ops:
                logger.debug(self, op.__str__())

if __name__ == "__main__":
    from pyscf.pbc import gto
    cell = gto.Cell()
    cell.atom = """
      Cu 1.000000     1.00000      0.0000
      O  0.000000     1.00000      0.0000
      O  1.000000     2.00000      0.0000
      Cu 1.000000     3.00000      0.0000
      O  1.000000     4.00000      0.0000
      O  2.000000     3.00000      0.0000
      Cu 3.000000     3.00000      0.0000
      O  4.000000     3.00000      0.0000
      O  3.000000     2.00000      0.0000
      Cu 3.000000     1.00000      0.0000
      O  3.000000     0.00000      0.0000
      O  2.000000     1.00000      0.0000
    """
    cell.a = [[4.0, 0., 0.], [0., 4.0, 0.], [0., 0., 16.0]]
    cell.verbose = 5
    cell.dimension = 3
    cell.magmom = [1., 1., -1., -1., 1., -1., 1., 1., -1., -1., 1., -1.]
    cell.build()
    sg = SpaceGroup(cell)
    #sg.backend = 'spglib'
    sg.build()
    print(sg.groupname['point_group_symbol'])
    print(sg.ops[0])
    print(hash(sg.ops[0]))
    sg.ops.reverse()
    print(sg.ops[0])
    sg.ops.sort()
    print(sg.ops[0])



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
density fitting MP2,  3-center integrals incore.
'''

import numpy as np
from pyscf import lib
from pyscf.lib import logger
from pyscf.mp import dfump2
from pyscf import __config__

WITH_T2 = getattr(__config__, 'mp_dfump2_with_t2', True)


def kernel(mp, mo_energy=None, mo_coeff=None, eris=None, with_t2=WITH_T2, verbose=None):

    log = logger.new_logger(mp, verbose)

    if eris is None: eris = mp.ao2mo()

    nocc, nvir = eris.nocc, eris.nvir
    split_mo_energy = mp.split_mo_energy()
    occ_energy = [x[1] for x in split_mo_energy]
    vir_energy = [x[2] for x in split_mo_energy]

    mem_avail = mp.max_memory - lib.current_memory()[0]

    if with_t2:
        t2 = (np.zeros((nocc[0],nocc[0],nvir[0],nvir[0]), dtype=eris.dtype),
              np.zeros((nocc[0],nocc[1],nvir[0],nvir[1]), dtype=eris.dtype),
              np.zeros((nocc[1],nocc[1],nvir[1],nvir[1]), dtype=eris.dtype))
        mem_avail -= sum([x.size for x in t2]) * eris.dsize / 1e6
    else:
        t2 = None

    if mem_avail < 0:
        log.error('Insufficient memory for holding t2 incore. Please rerun with `with_t2 = False`.')
        raise MemoryError

    cput1 = (logger.process_clock(), logger.perf_counter())

    emp2_ss = emp2_os = 0
    # same spin
    for s in [0,1]:
        s_t2 = 0 if s == 0 else 2
        moeoo = occ_energy[s][:,None] + occ_energy[s]
        moevv = lib.asarray(vir_energy[s][:,None] + vir_energy[s], order='C')
        for i in range(nocc[s]):
            ivL = eris.get_occ_blk(s,i,i+1)[0]
            for j in range(i+1):
                fac = 0.5 if i == j else 1

                if j == i:
                    jvL = ivL
                else:
                    jvL = eris.get_occ_blk(s,j,j+1)[0]

                vab = lib.dot(ivL, jvL.T)
                tab = np.conj(vab) / (moeoo[i,j] - moevv)
                ed =  lib.einsum('ab,ab->', vab, tab) * fac
                ex = -lib.einsum('ab,ba->', vab, tab) * fac
                emp2_ss += ed + ex

                if with_t2:
                    tab -= tab.T.conj()
                    t2[s_t2][i,j] = tab
                    if i != j:
                        t2[s_t2][j,i] = tab.T.conj()

            cput1 = log.timer_debug1('(sa,sb) = (%d,%d)  i-block [%d:%d]/%d' % (s,s,i,i+1,nocc[s]),
                                     *cput1)

    # opposite spin
    sa, sb = 0, 1
    moeoo = occ_energy[sa][:,None] + occ_energy[sb]
    moevv = lib.asarray(vir_energy[sa][:,None] + vir_energy[sb], order='C')
    for i in range(nocc[sa]):
        ivL = eris.get_occ_blk(sa,i,i+1)[0]
        for j in range(nocc[sb]):
            jvL = eris.get_occ_blk(sb,j,j+1)[0]

            vab = lib.dot(ivL, jvL.T)
            tab = np.conj(vab) / (moeoo[i,j] - moevv)
            ed =  lib.einsum('ab,ab->', vab, tab)
            emp2_os += ed

            if with_t2:
                t2[1][i,j] = tab

        cput1 = log.timer_debug1('(sa,sb) = (%d,%d)  i-block [%d:%d]/%d' % (sa,sb,i,i+1,nocc[sa]),
                                 *cput1)

    emp2_ss = emp2_ss.real
    emp2_os = emp2_os.real
    emp2 = lib.tag_array(emp2_ss+emp2_os, e_corr_ss=emp2_ss, e_corr_os=emp2_os)

    return emp2, t2


class DFUMP2(dfump2.DFUMP2):

    def init_amps(self, mo_energy=None, mo_coeff=None, eris=None, with_t2=WITH_T2):
        return kernel(self, mo_energy, mo_coeff, eris, with_t2)

UMP2 = DFUMP2

del (WITH_T2)


if __name__ == '__main__':
    from pyscf import scf
    from pyscf import gto, df
    mol = gto.Mole()
    mol.verbose = 0
    mol.atom = [
        [8 , (0. , 0.     , 0.)],
        [1 , (0. , -0.757 , 0.587)],
        [1 , (0. , 0.757  , 0.587)]]

    mol.basis = 'cc-pvdz'
    mol.charge = 1
    mol.spin = 1
    mol.build()
    mf = scf.UHF(mol).run()
    pt = DFUMP2(mf)
    emp2, t2 = pt.kernel()
    print(emp2 - -0.15321910988237386)

    pt.with_df = df.DF(mol)
    pt.with_df.auxbasis = 'weigend'
    emp2, t2 = pt.kernel()
    print(emp2 - -0.15345631939967935)

    mf = scf.density_fit(scf.UHF(mol), 'weigend')
    mf.kernel()
    pt = DFUMP2(mf)
    emp2, t2 = pt.kernel()
    print(emp2 - -0.15325911974166384)

    pt.with_df = df.DF(mol)
    pt.with_df.auxbasis = df.make_auxbasis(mol, mp2fit=True)
    emp2, t2 = pt.kernel()
    print(emp2 - -0.15302393704336487)

    pt.frozen = 2
    emp2, t2 = pt.kernel()
    print(emp2 - -0.09563606544882836)



#!/usr/bin/env python
# Copyright 2014-2023 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Matthew Hennefarth <mhennefarth@uchicago.edu>

import numpy as np
from scipy import linalg

from pyscf.lib import logger
from pyscf.fci import direct_spin1
from pyscf import __config__

from pyscf import mcpdft
from pyscf.mcpdft import _dms
from pyscf.mcpdft import chkfile


def weighted_average_densities(mc, ci=None, weights=None):
    """Compute the weighted average 1- and 2-electron CAS densities.
    1-electron CAS is returned as spin-separated.

    Args:
        mc : instance of class _PDFT

        ci : list of ndarrays of length nroots
            CI vectors should be from a converged CASSCF/CASCI calculation

        weights : ndarray of length nroots
            Weight for each state. If none, uses weights from SA-CASSCF
            calculation

    Returns:
        A tuple, the first is casdm1s and the second is casdm2 where they are
        weighted averages where the weights are given.
    """

    return _dms.make_weighted_casdm1s(
        mc, ci=ci, weights=weights
    ), _dms.make_weighted_casdm2(mc, ci=ci, weights=weights)


def get_lpdft_hconst(
    mc,
    E_ot,
    casdm1s_0,
    casdm2_0,
    hyb=1.0,
    ncas=None,
    ncore=None,
    veff1=None,
    veff2=None,
    mo_coeff=None,
):
    """Compute h_const for the L-PDFT Hamiltonian

    Args:
        mc : instance of class _PDFT

        E_ot : float
            On-top energy

        casdm1s_0 : ndarray of shape (2, ncas, ncas)
            Spin-separated 1-RDM in the active space generated from expansion
            density.

        casdm2_0 : ndarray of shape (ncas, ncas, ncas, ncas)
            Spin-summed 2-RDM in the active space generated from expansion
            density.

    Kwargs:
        hyb : float
            Hybridization constant (lambda term)

        ncas : float
            Number of active space MOs

        ncore: float
            Number of core MOs

        veff1 : ndarray of shape (nao, nao)
            1-body effective potential in the AO basis computed using the
            zeroth-order densities.

        veff2 : pyscf.mcscf.mc_ao2mo._ERIS instance
            Relevant 2-body effective potential in the MO basis.

    Returns:
        Constant term h_const for the expansion term.
    """

    if ncas is None:
        ncas = mc.ncas
    if ncore is None:
        ncore = mc.ncore
    if veff1 is None:
        veff1 = mc.veff1
    if veff2 is None:
        veff2 = mc.veff2
    if mo_coeff is None:
        mo_coeff = mc.mo_coeff

    nocc = ncore + ncas

    # Get the 1-RDM matrices
    casdm1_0 = casdm1s_0[0] + casdm1s_0[1]
    dm1s = _dms.casdm1s_to_dm1s(mc, casdm1s=casdm1s_0, mo_coeff=mo_coeff)
    dm1 = dm1s[0] + dm1s[1]

    # Coulomb interaction
    vj = mc._scf.get_j(dm=dm1)
    e_veff1_j = np.tensordot(veff1 + hyb * 0.5 * vj, dm1)

    # Deal with 2-electron on-top potential energy
    e_veff2 = veff2.energy_core
    e_veff2 += np.tensordot(veff2.vhf_c[ncore:nocc, ncore:nocc], casdm1_0)
    e_veff2 += 0.5 * np.tensordot(
        veff2.papa[ncore:nocc, :, ncore:nocc, :], casdm2_0, axes=4
    )

    # h_nuc + E_ot - 1/2 g_pqrs D_pq D_rs - V_pq D_pq - 1/2 v_pqrs d_pqrs
    energy_core = hyb * mc.energy_nuc() + E_ot - e_veff1_j - e_veff2
    return energy_core


def transformed_h1e_for_cas(
    mc, E_ot, casdm1s_0, casdm2_0, hyb=1.0, mo_coeff=None, ncas=None, ncore=None
):
    """Compute the CAS one-particle L-PDFT Hamiltonian

    Args:
        mc : instance of a _PDFT object

        E_ot : float
            On-top energy

        casdm1s_0 : ndarray of shape (2,ncas,ncas)
            Spin-separated 1-RDM in the active space generated from expansion
            density

        casdm2_0 : ndarray of shape (ncas,ncas,ncas,ncas)
            Spin-summed 2-RDM in the active space generated from expansion
            density

        hyb : float
            Hybridization constant (lambda term)

        mo_coeff : ndarray of shape (nao,nmo)
            A full set of molecular orbital coefficients. Taken from self if
            not provided.

        ncas : int
            Number of active space molecular orbitals

        ncore : int
            Number of core molecular orbitals

    Returns:
        A tuple, the first is the effective one-electron linear PDFT
        Hamiltonian defined in CAS space, the second is the modified core
        energy.
    """
    if mo_coeff is None:
        mo_coeff = mc.mo_coeff
    if ncas is None:
        ncas = mc.ncas
    if ncore is None:
        ncore = mc.ncore

    nocc = ncore + ncas
    mo_core = mo_coeff[:, :ncore]
    mo_cas = mo_coeff[:, ncore:nocc]

    # h_pq + V_pq + J_pq all in AO integrals
    hcore_eff = mc.get_lpdft_hcore_only(casdm1s_0, hyb=hyb, mo_coeff=mo_coeff,
                                        ncore=ncore, ncas=ncas)
    energy_core = mc.get_lpdft_hconst(E_ot, casdm1s_0, casdm2_0, hyb,
                                      mo_coeff=mo_coeff, ncore=ncore,
                                      ncas=ncas)

    if mo_core.size != 0:
        core_dm = np.dot(mo_core, mo_core.conj().T) * 2
        # This is precomputed in MRH's ERIS object
        energy_core += mc.veff2.energy_core
        energy_core += np.tensordot(core_dm, hcore_eff).real

    h1eff = mo_cas.conj().T @ hcore_eff @ mo_cas
    # Add in the 2-electron portion that acts as a 1-electron operator
    h1eff += mc.veff2.vhf_c[ncore:nocc, ncore:nocc]

    return h1eff, energy_core


def get_transformed_h2eff_for_cas(mc, ncore=None, ncas=None):
    """Compute the CAS two-particle linear PDFT Hamiltonian

    Args:
        ncore : int
            Number of core MOs

        ncas : int
            Number of active space MOs

    Returns:
        ndarray of shape (ncas,ncas,ncas,ncas) which contain v_vwxy
    """
    if ncore is None:
        ncore = mc.ncore
    if ncas is None:
        ncas = mc.ncas
    nocc = ncore + ncas
    return mc.veff2.papa[ncore:nocc, :, ncore:nocc, :]


def make_lpdft_ham_(mc, mo_coeff=None, ci=None, ot=None):
    """Compute the L-PDFT Hamiltonian

    Args:
        mo_coeff : ndarray of shape (nao, nmo)
            A full set of molecular orbital coefficients. Taken from self if
            not provided.

        ci : list of ndarrays of length nroots
            CI vectors should be from a converged CASSCF/CASCI calculation

        ot : an instance of on-top functional class - see otfnal.py

    Returns:
        lpdft_ham : ndarray of shape (nroots, nroots) or (nirreps, nroots, nroots)
            Linear approximation to the MC-PDFT energy expressed as a
            hamiltonian in the basis provided by the CI vectors. If
            StateAverageMix, then returns the block diagonal of the lpdft
            hamiltonian for each irrep.
    """

    if mo_coeff is None:
        mo_coeff = mc.mo_coeff
    if ci is None:
        ci = mc.ci
    if ot is None:
        ot = mc.otfnal

    ot.reset(mol=mc.mol)

    spin = abs(mc.nelecas[0] - mc.nelecas[1])
    omega, _, hyb = ot._numint.rsh_and_hybrid_coeff(ot.otxc, spin=spin)
    if abs(omega) > 1e-11:
        raise NotImplementedError("range-separated on-top functionals")
    if abs(hyb[0] - hyb[1]) > 1e-11:
        raise NotImplementedError(
            "hybrid functionals with different exchange, correlations components"
        )

    cas_hyb = hyb[0]

    ncas = mc.ncas
    casdm1s_0, casdm2_0 = mc.get_casdm12_0(ci=ci)

    mc.veff1, mc.veff2, E_ot = mc.get_pdft_veff(
        mo=mo_coeff,
        ci=ci,
        casdm1s=casdm1s_0,
        casdm2=casdm2_0,
        drop_mcwfn=True,
        incl_energy=True,
        ot=ot
    )

    # This is all standard procedure for generating the hamiltonian in PySCF
    h1, h0 = mc.get_h1lpdft(E_ot, casdm1s_0, casdm2_0, hyb=1.0 - cas_hyb, mo_coeff=mo_coeff)
    h2 = mc.get_h2lpdft()
    h2eff = direct_spin1.absorb_h1e(h1, h2, ncas, mc.nelecas, 0.5)

    def construct_ham_slice(solver, slice, nelecas):
        ci_irrep = ci[slice]
        if hasattr(solver, "orbsym"):
            solver.orbsym = mc.fcisolver.orbsym

        hc_all_irrep = [solver.contract_2e(h2eff, c, ncas, nelecas) for c in ci_irrep]
        lpdft_irrep = np.tensordot(ci_irrep, hc_all_irrep, axes=((1, 2), (1, 2)))
        diag_idx = np.diag_indices_from(lpdft_irrep)
        lpdft_irrep[diag_idx] += h0 + cas_hyb * mc.e_mcscf[slice]
        return lpdft_irrep

    if not isinstance(mc, _LPDFTMix):
        return construct_ham_slice(direct_spin1, slice(0, len(ci)), mc.nelecas)

    # We have a StateAverageMix Solver
    mc._irrep_slices = []
    start = 0
    for solver in mc.fcisolver.fcisolvers:
        end = start + solver.nroots
        mc._irrep_slices.append(slice(start, end))
        start = end

    return [
        construct_ham_slice(s, irrep, mc.fcisolver._get_nelec(s, mc.nelecas))
        for s, irrep in zip(mc.fcisolver.fcisolvers, mc._irrep_slices)
    ]


def kernel(mc, mo_coeff=None, ci0=None, ot=None, dump_chk=True, **kwargs):
    if ot is None:
        ot = mc.otfnal
    if mo_coeff is None:
        mo_coeff = mc.mo_coeff

    mc.optimize_mcscf_(mo_coeff=mo_coeff, ci0=ci0)
    ci_mcscf = mc.ci
    mc.lpdft_ham = mc.make_lpdft_ham_(ot=ot)
    logger.debug(mc, f"L-PDFT Hamiltonian in MC-SCF Basis:\n{mc.get_lpdft_ham()}")

    if hasattr(mc, "_irrep_slices"):
        e_states, si_pdft = zip(*map(mc._eig_si, mc.lpdft_ham))
        e_states = np.concatenate(e_states)
        si_pdft = linalg.block_diag(*si_pdft)

    else:
        e_states, si_pdft = mc._eig_si(mc.lpdft_ham)

    mc.e_states = e_states
    mc.si_pdft = si_pdft

    logger.debug(mc, f"L-PDFT SI:\n{mc.si_pdft}")

    e_tot = np.dot(e_states, mc.weights)
    ci = mc._get_ci_adiabats(ci_mcscf)

    mc.e_tot = e_tot
    mc.ci = ci

    if dump_chk:
        mc.dump_chk(locals())

    return (mc.e_tot, mc.e_mcscf, mc.e_cas, mc.ci, mc.mo_coeff, mc.mo_energy)


class _LPDFT(mcpdft.MultiStateMCPDFTSolver):
    """Linerized PDFT

    Saved Results

        e_tot : float
            Weighted-average L-PDFT final energy
        e_states : ndarray of shape (nroots)
            L-PDFT final energies of the adiabatic states
        ci : list of length (nroots) of ndarrays
            CI vectors in the optimized adiabatic basis of L-PDFT
        si_pdft : ndarray of shape (nroots, nroots)
            Expansion coefficients of the L-PDFT adiabats in terms of the
            optimized MC-SCF adiabats
        e_mcscf : ndarray of shape (nroots)
            Energies of the MC-SCF adiabatic states
        lpdft_ham : ndarray of shape (nroots, nroots)
            L-PDFT Hamiltonian in the MC-SCF adiabatic basis
        veff1 : ndarray of shape (nao, nao)
            1-body effective potential in the AO basis computed using the
            zeroth-order densities.
        veff2 : pyscf.mcscf.mc_ao2mo._ERIS instance
            Relevant 2-body effective potential in the MO basis.
    """

    # chk_veff = getattr(__config__, 'mcpdft_lpdft_chk_veff', False)

    def __init__(self, mc):
        self.__dict__.update(mc.__dict__)
        keys = set(("lpdft_ham", "si_pdft", "veff1", "veff2"))
        self.lpdft_ham = None
        self.si_pdft = None
        self.veff1 = None
        self.veff2 = None
        self._e_states = None
        self._keys = set(self.__dict__.keys()).union(keys)

    @property
    def e_states(self):
        if self._in_mcscf_env:
            return self.fcisolver.e_states

        else:
            return self._e_states

    @e_states.setter
    def e_states(self, x):
        self._e_states = x

    make_lpdft_ham_ = make_lpdft_ham_
    make_lpdft_ham_.__doc__ = make_lpdft_ham_.__doc__

    get_lpdft_hconst = get_lpdft_hconst
    get_lpdft_hconst.__doc__ = get_lpdft_hconst.__doc__

    get_h1lpdft = transformed_h1e_for_cas
    get_h1lpdft.__doc__ = transformed_h1e_for_cas.__doc__

    get_h2lpdft = get_transformed_h2eff_for_cas
    get_h2lpdft.__doc__ = get_transformed_h2eff_for_cas.__doc__

    get_casdm12_0 = weighted_average_densities
    get_casdm12_0.__doc__ = weighted_average_densities.__doc__

    def get_lpdft_diag(self):
        """Diagonal elements of the L-PDFT Hamiltonian matrix
            (H_00^L-PDFT, H_11^L-PDFT, H_22^L-PDFT, ...)

        Returns:
            lpdft_diag : ndarray of shape (nroots)
                Contains the linear approximation to the MC-PDFT energy. These
                are also the diagonal elements of the L-PDFT Hamiltonian
                matrix.
        """
        return np.diagonal(self.lpdft_ham).copy()

    def get_lpdft_ham(self):
        """The L-PDFT effective Hamiltonian matrix

        Returns:
            lpdft_ham : ndarray of shape (nroots, nroots)
                Contains L-PDFT Hamiltonian elements on the off-diagonals
                and PDFT approx energies on the diagonals
        """
        return self.lpdft_ham

    def kernel(self, mo_coeff=None, ci0=None, ot=None, verbose=None, dump_chk=True):
        """
        Returns:
            6 elements, they are
            total energy,
            the MCSCF energies,
            the active space CI energy,
            the active space FCI wave function coefficients,
            the MCSCF canonical orbital coefficients,
            the MCSCF canonical orbital energies

        They are attributes of the QLPDFT object, which can be accessed by
        .e_tot, .e_mcscf, .e_cas, .ci, .mo_coeff, .mo_energy
        """
        if ot is None:
            ot = self.otfnal
        ot.reset(mol=self.mol)  # scanner mode safety

        if mo_coeff is None:
            mo_coeff = self.mo_coeff
        else:
            self.mo_coeff = mo_coeff

        log = logger.new_logger(self, verbose)

        if ci0 is None and isinstance(getattr(self, "ci", None), list):
            ci0 = [c.copy() for c in self.ci]

        kernel(self, mo_coeff, ci0, ot=ot, verbose=log, dump_chk=dump_chk)
        self._finalize_lin()
        return (
            self.e_tot,
            self.e_mcscf,
            self.e_cas,
            self.ci,
            self.mo_coeff,
            self.mo_energy,
        )

    def _finalize_lin(self):
        log = logger.Logger(self.stdout, self.verbose)
        nroots = len(self.e_states)
        log.note("%s (final) states:", self.__class__.__name__)
        if log.verbose >= logger.NOTE and getattr(self.fcisolver, "spin_square", None):
            ss = self.fcisolver.states_spin_square(self.ci, self.ncas, self.nelecas)[0]

            for i in range(nroots):
                log.note(
                    "  State %d weight %g  ELPDFT = %.15g  S^2 = %.7f",
                    i,
                    self.weights[i],
                    self.e_states[i],
                    ss[i],
                )

        else:
            for i in range(nroots):
                log.note(
                    "  State %d weight %g  ELPDFT = %.15g",
                    i,
                    self.weights[i],
                    self.e_states[i],
                )

    def _get_ci_adiabats(self, ci_mcscf):
        """Get the CI vertors in eigenbasis of L-PDFT Hamiltonian

        Kwargs:
            ci : list of length nroots
                MC-SCF ci vectors; defaults to self.ci_mcscf

        Returns:
            ci : list of length nroots
                CI vectors in basis of L-PDFT Hamiltonian eigenvectors
        """
        return list(np.tensordot(self.si_pdft.T, np.asarray(ci_mcscf), axes=1))

    def _eig_si(self, ham):
        return linalg.eigh(ham)

    def get_lpdft_hcore_only(self, casdm1s_0, hyb=1.0, mo_coeff=None, ncore=None, ncas=None):
        """
        Returns the lpdft hcore AO integrals weighted by the
        hybridization factor. Excludes the MC-SCF (wfn) component.
        """

        dm1s = _dms.casdm1s_to_dm1s(self, casdm1s=casdm1s_0, mo_coeff=mo_coeff,
                                    ncore=ncore, ncas=ncas)
        dm1 = dm1s[0] + dm1s[1]
        v_j = self._scf.get_j(dm=dm1)
        return hyb * self.get_hcore() + self.veff1 + hyb * v_j

    def get_lpdft_hcore(self, casdm1s_0=None, mo_coeff=None, ncore=None, ncas=None):
        """
        Returns the full lpdft hcore AO integrals. Includes the MC-SCF
        (wfn) component for hybrid functionals.
        """
        if casdm1s_0 is None:
            casdm1s_0 = self.get_casdm12_0()[0]

        spin = abs(self.nelecas[0] - self.nelecas[1])
        cas_hyb = self.otfnal._numint.rsh_and_hybrid_coeff(self.otfnal.otxc, spin=spin)[
            2
        ]
        hyb = 1.0 - cas_hyb[0]

        return cas_hyb[0] * self.get_hcore() + self.get_lpdft_hcore_only(
            casdm1s_0, hyb=hyb, mo_coeff=mo_coeff, ncore=ncore, ncas=ncas
        )

    def nuc_grad_method(self, state=None):
        from pyscf.mcscf import mc1step
        from pyscf.mcscf.df import _DFCASSCF

        if not isinstance(self, mc1step.CASSCF):
            raise NotImplementedError("CASCI-based LPDFT nuclear gradients")
        elif getattr(self, "frozen", None) is not None:
            raise NotImplementedError("LPDFT nuclear gradients with frozen orbitals")
        elif isinstance(self, _DFCASSCF):
            from pyscf.df.grad.lpdft import Gradients
        else:
            from pyscf.grad.lpdft import Gradients

        return Gradients(self, state=state)

    def dump_chk(self, envs):
        if self.chkfile is None:
            return self

        if self._in_mcscf_env:
            self._mc_class.dump_chk(self, envs)

        else:
            ci = None
            if self.chk_ci:
                ci = envs["ci"]

            # if self.chk_veff:
            # veff1 = self.veff1
            # veff2 = self.veff2

            chkfile.dump_lpdft(
                self,
                chkfile=self.chkfile,
                key="pdft",
                e_tot=envs["e_tot"],
                e_states=envs["e_states"],
                e_mcscf=self.e_mcscf,
                ci=ci,
            )

        return self


class _LPDFTMix(_LPDFT):
    """State Averaged Mixed Linerized PDFT

    Saved Results

        e_tot : float
            Weighted-average L-PDFT final energy
        e_states : ndarray of shape (nroots)
            L-PDFT final energies of the adiabatic states
        ci : list of length (nroots) of ndarrays
            CI vectors in the optimized adiabatic basis of MC-SCF. Related to the
            L-PDFT adiabat CI vectors by the expansion coefficients ``si_pdft''.
        si_pdft : ndarray of shape (nroots, nroots)
            Expansion coefficients of the L-PDFT adiabats in terms of the optimized
            MC-SCF adiabats
        e_mcscf : ndarray of shape (nroots)
            Energies of the MC-SCF adiabatic states
        lpdft_ham : list of ndarray of shape (nirreps, nroots, nroots)
            L-PDFT Hamiltonian in the MC-SCF adiabatic basis within each irrep
    """

    def __init__(self, mc):
        super().__init__(mc)
        # Holds the irrep slices for when we need to index into various quantities
        self._irrep_slices = None

    def get_lpdft_diag(self):
        """Diagonal elements of the L-PDFT Hamiltonian matrix
            (H_00^L-PDFT, H_11^L-PDFT, H_22^L-PDFT, ...)

        Returns:
            lpdft_diag : ndarray of shape (nroots)
                Contains the linear approximation to the MC-PDFT energy. These
                are also the diagonal elements of the L-PDFT Hamiltonian
                matrix.
        """
        return np.concatenate(
            [np.diagonal(irrep_ham).copy() for irrep_ham in self.lpdft_ham]
        )

    def get_lpdft_ham(self):
        """The L-PDFT effective Hamiltonian matrix

        Returns:
            lpdft_ham : ndarray of shape (nroots, nroots)
                Contains L-PDFT Hamiltonian elements on the off-diagonals
                and PDFT approx energies on the diagonals
        """
        return linalg.block_diag(*self.lpdft_ham)

    def _get_ci_adiabats(self, ci_mcscf):
        """Get the CI vertors in eigenbasis of L-PDFT Hamiltonian

        ci : list of length nroots
            MC-SCF ci vectors

        Returns:
            ci : list of length nroots
                CI vectors in basis of L-PDFT Hamiltonian eigenvectors
        """
        adiabat_ci = [
            np.tensordot(
                self.si_pdft[irrep_slice, irrep_slice],
                np.asarray(ci_mcscf[irrep_slice]),
                axes=1,
            )
            for irrep_slice in self._irrep_slices
        ]
        # Flattens it
        return [c for ci_irrep in adiabat_ci for c in ci_irrep]

    def nuc_grad_method(self, state=None):
        raise NotImplementedError("MultiState Mix LPDFT nuclear gradients")


def linear_multi_state(mc, weights=(0.5, 0.5), **kwargs):
    """Build linearized multi-state MC-PDFT method object

    Args:
        mc : instance of class _PDFT

    Kwargs:
        weights : sequence of floats

    Returns:
        si : instance of class _LPDFT
    """
    from pyscf.mcscf.addons import StateAverageMCSCFSolver, StateAverageMixFCISolver

    if isinstance(mc, mcpdft.MultiStateMCPDFTSolver):
        raise RuntimeError("already a multi-state PDFT solver")

    if isinstance(mc.fcisolver, StateAverageMixFCISolver):
        raise RuntimeError("state-average mix type")

    if not isinstance(mc, StateAverageMCSCFSolver):
        base_name = mc.__class__.__name__
        mc = mc.state_average(weights=weights, **kwargs)

    else:
        base_name = mc.__class__.bases__[0].__name__

    mcbase_class = mc.__class__

    class LPDFT(_LPDFT, mcbase_class):
        pass

    LPDFT.__name__ = "LIN" + base_name
    return LPDFT(mc)


def linear_multi_state_mix(mc, fcisolvers, weights=(0.5, 0.5), **kwargs):
    """Build SA Mix linearized multi-state MC-PDFT method object

    Args:
        mc : instance of class _PDFT

        fcisolvers : fcisolvers to construct StateAverageMixSolver with

    Kwargs:
        weights : sequence of floats

    Returns:
        si : instance of class _LPDFT
    """
    from pyscf.mcscf.addons import StateAverageMCSCFSolver, StateAverageMixFCISolver

    if isinstance(mc, mcpdft.MultiStateMCPDFTSolver):
        raise RuntimeError("already a multi-state PDFT solver")

    if not isinstance(mc, StateAverageMCSCFSolver):
        base_name = mc.__class__.__name__
        mc = mc.state_average_mix(fcisolvers, weights=weights, **kwargs)

    elif not isinstance(mc.fcisolver, StateAverageMixFCISolver):
        raise RuntimeError("already a StateAverageMCSCF solver")

    else:
        base_name = mc.__class__.bases__[0].__name__

    mcbase_class = mc.__class__

    class LPDFT(_LPDFTMix, mcbase_class):
        pass

    LPDFT.__name__ = "LIN" + base_name
    return LPDFT(mc)


if __name__ == "__main__":
    from pyscf import gto, scf
    from pyscf import mcpdft
    from pyscf.csf_fci import csf_solver

    mol = gto.M(
        atom="""H 0 0 0
                       H 1.5 0 0""",
        basis="sto-3g",
        verbose=5,
        spin=0,
        unit="AU",
        symmetry=True,
    )

    mf = scf.RHF(mol).run()

    mc = mcpdft.CASSCF(mf, "tPBE", 2, 2, grids_level=6)
    mc.fcisolver = csf_solver(mol, smult=1)

    N_STATES = 2

    mc = mc.state_average(
        [
            1.0 / float(N_STATES),
        ]
        * N_STATES
    )

    sc = linear_multi_state(mc)
    sc.kernel()



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#         Timothy Berkelbach <tim.berkelbach@gmail.com>
#

'''
RHF-QCISD(T) for real integrals
'''


import ctypes
import numpy
from pyscf import lib
from pyscf.lib import logger
from pyscf.cc import _ccsd
from pyscf.cc.ccsd_t import _sort_eri, _sort_t2_vooo_

# t3 as ijkabc

# JCP 94, 442 (1991); DOI:10.1063/1.460359.  Error in Eq (1), should be [ia] >= [jb] >= [kc]
def kernel(mycc, eris, t1=None, t2=None, verbose=logger.NOTE):
    cpu1 = cpu0 = (logger.process_clock(), logger.perf_counter())
    log = logger.new_logger(mycc, verbose)
    if t1 is None: t1 = mycc.t1
    if t2 is None: t2 = mycc.t2

    nocc, nvir = t1.shape
    nmo = nocc + nvir

    dtype = numpy.result_type(t1, t2, eris.ovoo.dtype)
    if mycc.incore_complete:
        ftmp = None
        eris_vvop = numpy.zeros((nvir,nvir,nocc,nmo), dtype)
    else:
        ftmp = lib.H5TmpFile()
        eris_vvop = ftmp.create_dataset('vvop', (nvir,nvir,nocc,nmo), dtype)

    orbsym = _sort_eri(mycc, eris, nocc, nvir, eris_vvop, log)

    mo_energy, t1T, t2T, vooo, fvo, restore_t2_inplace = \
            _sort_t2_vooo_(mycc, orbsym, t1, t2, eris)
    cpu1 = log.timer_debug1('QCISD(T) sort_eri', *cpu1)

    cpu2 = list(cpu1)
    orbsym = numpy.hstack((numpy.sort(orbsym[:nocc]),numpy.sort(orbsym[nocc:])))
    o_ir_loc = numpy.append(0, numpy.cumsum(numpy.bincount(orbsym[:nocc], minlength=8)))
    v_ir_loc = numpy.append(0, numpy.cumsum(numpy.bincount(orbsym[nocc:], minlength=8)))
    o_sym = orbsym[:nocc]
    oo_sym = (o_sym[:,None] ^ o_sym).ravel()
    oo_ir_loc = numpy.append(0, numpy.cumsum(numpy.bincount(oo_sym, minlength=8)))
    nirrep = max(oo_sym) + 1

    orbsym   = orbsym.astype(numpy.int32)
    o_ir_loc = o_ir_loc.astype(numpy.int32)
    v_ir_loc = v_ir_loc.astype(numpy.int32)
    oo_ir_loc = oo_ir_loc.astype(numpy.int32)
    if dtype == numpy.complex128:
        drv = _ccsd.libcc.QCIsd_t_zcontract
    else:
        drv = _ccsd.libcc.QCIsd_t_contract
    et_sum = numpy.zeros(1, dtype=dtype)
    def contract(a0, a1, b0, b1, cache):
        cache_row_a, cache_col_a, cache_row_b, cache_col_b = cache
        drv(et_sum.ctypes.data_as(ctypes.c_void_p),
            mo_energy.ctypes.data_as(ctypes.c_void_p),
            t1T.ctypes.data_as(ctypes.c_void_p),
            t2T.ctypes.data_as(ctypes.c_void_p),
            vooo.ctypes.data_as(ctypes.c_void_p),
            fvo.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(nocc), ctypes.c_int(nvir),
            ctypes.c_int(a0), ctypes.c_int(a1),
            ctypes.c_int(b0), ctypes.c_int(b1),
            ctypes.c_int(nirrep),
            o_ir_loc.ctypes.data_as(ctypes.c_void_p),
            v_ir_loc.ctypes.data_as(ctypes.c_void_p),
            oo_ir_loc.ctypes.data_as(ctypes.c_void_p),
            orbsym.ctypes.data_as(ctypes.c_void_p),
            cache_row_a.ctypes.data_as(ctypes.c_void_p),
            cache_col_a.ctypes.data_as(ctypes.c_void_p),
            cache_row_b.ctypes.data_as(ctypes.c_void_p),
            cache_col_b.ctypes.data_as(ctypes.c_void_p))
        cpu2[:] = log.timer_debug1('contract %d:%d,%d:%d'%(a0,a1,b0,b1), *cpu2)

    # The rest 20% memory for cache b
    mem_now = lib.current_memory()[0]
    max_memory = max(0, mycc.max_memory - mem_now)
    bufsize = (max_memory*.5e6/8-nocc**3*3*lib.num_threads())/(nocc*nmo)  #*.5 for async_io
    bufsize *= .5  #*.5 upper triangular part is loaded
    bufsize *= .8  #*.8 for [a0:a1]/[b0:b1] partition
    bufsize = max(8, bufsize)
    log.debug('max_memory %d MB (%d MB in use)', max_memory, mem_now)
    with lib.call_in_background(contract, sync=not mycc.async_io) as async_contract:
        for a0, a1 in reversed(list(lib.prange_tril(0, nvir, bufsize))):
            cache_row_a = numpy.asarray(eris_vvop[a0:a1,:a1], order='C')
            if a0 == 0:
                cache_col_a = cache_row_a
            else:
                cache_col_a = numpy.asarray(eris_vvop[:a0,a0:a1], order='C')
            async_contract(a0, a1, a0, a1, (cache_row_a,cache_col_a,
                                            cache_row_a,cache_col_a))

            for b0, b1 in lib.prange_tril(0, a0, bufsize/8):
                cache_row_b = numpy.asarray(eris_vvop[b0:b1,:b1], order='C')
                if b0 == 0:
                    cache_col_b = cache_row_b
                else:
                    cache_col_b = numpy.asarray(eris_vvop[:b0,b0:b1], order='C')
                async_contract(a0, a1, b0, b1, (cache_row_a,cache_col_a,
                                                cache_row_b,cache_col_b))

    t2 = restore_t2_inplace(t2T)
    et_sum *= 2
    if abs(et_sum[0].imag) > 1e-4:
        logger.warn(mycc, 'Non-zero imaginary part of QCISD(T) energy was found %s',
                    et_sum[0])
    et = et_sum[0].real
    log.timer('QCISD(T)', *cpu0)
    log.note('QCISD(T) correction = %.15g', et)
    return et



#!/usr/bin/env python
# Copyright 2014-2019 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Non-relativistic unrestricted Kohn-Sham for periodic systems at a single k-point

See Also:
    pyscf.pbc.dft.krks.py : Non-relativistic Restricted Kohn-Sham for periodic
                            systems with k-point sampling
'''


import numpy
import pyscf.dft
from pyscf import lib
from pyscf.pbc.scf import uhf as pbcuhf
from pyscf.lib import logger
from pyscf.dft import uks as mol_uks
from pyscf.pbc.dft import gen_grid
from pyscf.pbc.dft import rks
from pyscf.pbc.dft import multigrid
from pyscf import __config__

get_rho = rks.get_rho

def get_veff(ks, cell=None, dm=None, dm_last=0, vhf_last=0, hermi=1,
             kpt=None, kpts_band=None):
    '''Coulomb + XC functional for UKS.  See pyscf/pbc/dft/uks.py
    :func:`get_veff` fore more details.
    '''
    if cell is None: cell = ks.cell
    if dm is None: dm = ks.make_rdm1()
    if kpt is None: kpt = ks.kpt
    t0 = (logger.process_clock(), logger.perf_counter())

    if not isinstance(dm, numpy.ndarray):
        dm = numpy.asarray(dm)
    if dm.ndim == 2:  # RHF DM
        dm = numpy.asarray((dm*.5,dm*.5))

    ni = ks._numint
    if isinstance(ni, multigrid.MultiGridNumInt):
        if ks.do_nlc():
            raise NotImplementedError(f'MultiGrid for NLC functional {ks.xc} + {ks.nlc}')
        j_in_xc = ni.xc_with_j
    else:
        ks.initialize_grids(cell, dm, kpt)
        j_in_xc = False

    max_memory = ks.max_memory - lib.current_memory()[0]
    n, exc, vxc = ni.nr_uks(cell, ks.grids, ks.xc, dm, 0, hermi,
                            kpt, kpts_band, max_memory=max_memory)
    logger.info(ks, 'nelec by numeric integration = %s', n)
    if ks.do_nlc():
        if ni.libxc.is_nlc(ks.xc):
            xc = ks.xc
        else:
            assert ni.libxc.is_nlc(ks.nlc)
            xc = ks.nlc
        n, enlc, vnlc = ni.nr_nlc_vxc(cell, ks.nlcgrids, xc, dm[0]+dm[1],
                                      0, hermi, kpt, max_memory=max_memory)
        exc += enlc
        vxc += vnlc
        logger.info(ks, 'nelec with nlc grids = %s', n)
    t0 = logger.timer(ks, 'vxc', *t0)

    ground_state = kpts_band is None
    vj, vk = rks._get_jk(ks, cell, dm, hermi, kpt, kpts_band, with_j=not j_in_xc)
    if j_in_xc:
        ecoul = vxc.ecoul
    else:
        vj = vj[0] + vj[1]
        vxc += vj
        ecoul = None
        if ground_state:
            ecoul = numpy.einsum('nij,ji->', dm, vj).real * .5
    if ni.libxc.is_hybrid_xc(ks.xc):
        vxc -= vk
        if ground_state:
            exc -= numpy.einsum('nij,nji->', dm, vk).real * .5
    vxc = lib.tag_array(vxc, ecoul=ecoul, exc=exc, vj=None, vk=None)
    logger.timer(ks, 'veff', *t0)
    return vxc

def gen_response(mf, mo_coeff=None, mo_occ=None,
                 with_j=True, hermi=0, max_memory=None, with_nlc=True):
    if mo_coeff is None: mo_coeff = mf.mo_coeff
    if mo_occ is None: mo_occ = mf.mo_occ
    cell = mf.cell
    kpt = mf.kpt
    ni = mf._numint
    hybrid = ni.libxc.is_hybrid_xc(mf.xc)
    j_in_xc = getattr(ni, 'xc_with_j', False)

    if with_nlc and mf.do_nlc():
        raise NotImplementedError

    rho0, vxc, fxc = ni.cache_xc_kernel(cell, mf.grids, mf.xc,
                                        mo_coeff, mo_occ, 1, kpt)
    dm0 = None

    if max_memory is None:
        mem_now = lib.current_memory()[0]
        max_memory = max(2000, mf.max_memory*.8-mem_now)

    def vind(dm1):
        if hermi == 2:
            v1 = numpy.zeros_like(dm1)
        else:
            v1 = ni.nr_uks_fxc(cell, mf.grids, mf.xc, dm0, dm1, 0, hermi,
                               rho0, vxc, fxc, kpt, max_memory=max_memory)
        vj, vk = rks._get_jk(mf, cell, dm1, hermi, kpt, with_j=not j_in_xc)
        if not j_in_xc:
            v1 += vj[0] + vj[1]
        if hybrid:
            v1 -= vk
        return v1
    return vind

class UKS(rks.KohnShamDFT, pbcuhf.UHF):
    '''UKS class adapted for PBCs.

    This is a literal duplication of the molecular UKS class with some `mol`
    variables replaced by `cell`.
    '''

    get_rho = get_rho
    get_vsap = mol_uks.UKS.get_vsap
    init_guess_by_vsap = mol_uks.UKS.init_guess_by_vsap
    get_veff = get_veff
    energy_elec = pyscf.dft.uks.energy_elec
    gen_response = gen_response

    def __init__(self, cell, kpt=numpy.zeros(3), xc='LDA,VWN',
                 exxdiv=getattr(__config__, 'pbc_scf_SCF_exxdiv', 'ewald')):
        pbcuhf.UHF.__init__(self, cell, kpt, exxdiv=exxdiv)
        rks.KohnShamDFT.__init__(self, xc)

    def dump_flags(self, verbose=None):
        pbcuhf.UHF.dump_flags(self, verbose)
        rks.KohnShamDFT.dump_flags(self, verbose)
        return self

    def to_hf(self):
        '''Convert to UHF object.'''
        from pyscf.pbc import scf
        return self._transfer_attrs_(scf.UHF(self.cell, self.kpt))

    def Gradients(self):
        from pyscf.pbc.grad import uks
        from pyscf.pbc.dft.multigrid import MultiGridNumInt2
        if not isinstance(self._numint, MultiGridNumInt2):
            raise NotImplementedError('pbc-UKS must be computed with MultiGridNumInt2')
        return uks.Gradients(self)

    multigrid_numint = rks.RKS.multigrid_numint

    to_gpu = lib.to_gpu



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Extension to numpy and scipy
'''

import ctypes
import math
import numpy
from pyscf.lib import misc
from numpy import asarray  # For backward compatibility

EINSUM_MAX_SIZE = getattr(misc.__config__, 'lib_einsum_max_size', 2000)

try:
    # Import tblis before libnp_helper to avoid potential dl-loading conflicts
    from pyscf import tblis_einsum
    FOUND_TBLIS = True
except (ImportError, OSError):
    FOUND_TBLIS = False

_np_helper = misc.load_library('libnp_helper')

BLOCK_DIM = 192
PLAIN = 0
HERMITIAN = 1
ANTIHERMI = 2
SYMMETRIC = 3

LeviCivita = numpy.zeros((3,3,3))
LeviCivita[0,1,2] = LeviCivita[1,2,0] = LeviCivita[2,0,1] = 1
LeviCivita[0,2,1] = LeviCivita[2,1,0] = LeviCivita[1,0,2] = -1

PauliMatrices = numpy.array([[[0., 1.],
                              [1., 0.]],  # x
                             [[0.,-1j],
                              [1j, 0.]],  # y
                             [[1., 0.],
                              [0.,-1.]]]) # z

if hasattr(numpy, 'einsum_path'):
    _einsum_path = numpy.einsum_path
else:
    def _einsum_path(subscripts, *operands, **kwargs):
        #indices  = re.split(',|->', subscripts)
        #indices_in = indices[:-1]
        #idx_final = indices[-1]
        if '->' in subscripts:
            indices_in, idx_final = subscripts.split('->')
            indices_in = indices_in.split(',')
            # indices = indices_in + [idx_final]
        else:
            idx_final = ''
            indices_in = subscripts.split('->')[0].split(',')

        if len(indices_in) <= 2:
            idx_removed = set(indices_in[0]).intersection(set(indices_in[1]))
            einsum_str = indices_in[1] + ',' + indices_in[0] + '->' + idx_final
            return operands, [((1,0), idx_removed, einsum_str, idx_final)]

        input_sets = [set(x) for x in indices_in]
        n_shared_max = 0
        for i in range(len(indices_in)):
            for j in range(i):
                tmp = input_sets[i].intersection(input_sets[j])
                n_shared_indices = len(tmp)
                if n_shared_indices > n_shared_max:
                    n_shared_max = n_shared_indices
                    idx_removed = tmp
                    a,b = i,j

        idxA = indices_in.pop(a)
        idxB = indices_in.pop(b)
        rest_idx = ''.join(indices_in) + idx_final
        idx_out = input_sets[a].union(input_sets[b])
        idx_out = ''.join(idx_out.intersection(set(rest_idx)))

        indices_in.append(idx_out)
        einsum_str = idxA + ',' + idxB + '->' + idx_out
        einsum_args = _einsum_path(','.join(indices_in)+'->'+idx_final)[1]
        einsum_args.insert(0, ((a, b), idx_removed, einsum_str, indices_in))
        return operands, einsum_args

_numpy_einsum = numpy.einsum
def _contract(subscripts, *tensors, **kwargs):
    idx_str = subscripts.replace(' ','')
    A, B = tensors
    # Call numpy.asarray because A or B may be HDF5 Datasets
    A = numpy.asarray(A)
    B = numpy.asarray(B)

    # small problem size
    if A.size < EINSUM_MAX_SIZE or B.size < EINSUM_MAX_SIZE:
        return _numpy_einsum(idx_str, A, B)

    C_dtype = numpy.result_type(A, B)
    if FOUND_TBLIS and C_dtype == numpy.double:
        # tblis is slow for complex type
        return tblis_einsum.contract(idx_str, A, B, **kwargs)

    indices  = idx_str.replace(',', '').replace('->', '')
    if '->' not in idx_str or any(indices.count(x) != 2 for x in set(indices)):
        return _numpy_einsum(idx_str, A, B)

    # Split the strings into a list of idx char's
    idxA, idxBC = idx_str.split(',')
    idxB, idxC = idxBC.split('->')
    assert len(idxA) == A.ndim
    assert len(idxB) == B.ndim

    uniq_idxA = set(idxA)
    uniq_idxB = set(idxB)
    # Find the shared indices being summed over
    shared_idxAB = uniq_idxA.intersection(uniq_idxB)

    if ((not shared_idxAB) or  # Indices must overlap
        # one operand is a subset of the other one (e.g. 'ijkl,jk->il')
        uniq_idxA == shared_idxAB or uniq_idxB == shared_idxAB or
        # repeated indices (e.g. 'iijk,kl->jl')
        len(idxA) != len(uniq_idxA) or len(idxB) != len(uniq_idxB)):
        return _numpy_einsum(idx_str, A, B)

    DEBUG = kwargs.get('DEBUG', False)

    if DEBUG:
        print("*** Einsum for", idx_str)
        print(" idxA =", idxA)
        print(" idxB =", idxB)
        print(" idxC =", idxC)

    # Get the range for each index and put it in a dictionary
    rangeA = dict(zip(idxA, A.shape))
    rangeB = dict(zip(idxB, B.shape))
    #rangeC = dict(zip(idxC, C.shape))
    if DEBUG:
        print("rangeA =", rangeA)
        print("rangeB =", rangeB)

    idxAt = list(idxA)
    idxBt = list(idxB)
    inner_shape = 1
    insert_B_loc = 0
    shared_idxAB = sorted(shared_idxAB)
    for n in shared_idxAB:
        if rangeA[n] != rangeB[n]:
            err = ('ERROR: In index string %s, the range of index %s is '
                   'different in A (%d) and B (%d)' %
                   (idx_str, n, rangeA[n], rangeB[n]))
            raise ValueError(err)

        # Bring idx all the way to the right for A
        # and to the left (but preserve order) for B
        idxA_n = idxAt.index(n)
        idxAt.insert(len(idxAt)-1, idxAt.pop(idxA_n))

        idxB_n = idxBt.index(n)
        idxBt.insert(insert_B_loc, idxBt.pop(idxB_n))
        insert_B_loc += 1

        inner_shape *= rangeA[n]

    if DEBUG:
        print("shared_idxAB =", shared_idxAB)
        print("inner_shape =", inner_shape)

    # Transpose the tensors into the proper order and reshape into matrices
    new_orderA = [idxA.index(idx) for idx in idxAt]
    new_orderB = [idxB.index(idx) for idx in idxBt]

    if DEBUG:
        print("Transposing A as", new_orderA)
        print("Transposing B as", new_orderB)
        print("Reshaping A as (-1,", inner_shape, ")")
        print("Reshaping B as (", inner_shape, ",-1)")

    shapeCt = []
    idxCt = []
    for idx in idxAt:
        if idx in shared_idxAB:
            break
        shapeCt.append(rangeA[idx])
        idxCt.append(idx)
    for idx in idxBt:
        if idx in shared_idxAB:
            continue
        shapeCt.append(rangeB[idx])
        idxCt.append(idx)
    new_orderCt = [idxCt.index(idx) for idx in idxC]

    if A.size == 0 or B.size == 0:
        shapeCt = [shapeCt[i] for i in new_orderCt]
        return numpy.zeros(shapeCt, dtype=C_dtype)

    At = A.transpose(new_orderA)
    Bt = B.transpose(new_orderB)

    if At.flags.f_contiguous:
        At = numpy.asarray(At.reshape(-1,inner_shape), order='F')
    else:
        At = numpy.asarray(At.reshape(-1,inner_shape), order='C')
    if Bt.flags.f_contiguous:
        Bt = numpy.asarray(Bt.reshape(inner_shape,-1), order='F')
    else:
        Bt = numpy.asarray(Bt.reshape(inner_shape,-1), order='C')

    return dot(At,Bt).reshape(shapeCt, order='A').transpose(new_orderCt)

def einsum(subscripts, *tensors, **kwargs):
    '''Perform a more efficient einsum via reshaping to a matrix multiply.

    Current differences compared to numpy.einsum:
    This assumes that each repeated index is actually summed (i.e. no 'i,i->i')
    and appears only twice (i.e. no 'ij,ik,il->jkl'). The output indices must
    be explicitly specified (i.e. 'ij,j->i' and not 'ij,j').
    '''
    contract = kwargs.pop('_contract', _contract)

    subscripts = subscripts.replace(' ','')
    if len(tensors) <= 1 or '...' in subscripts:
        out = _numpy_einsum(subscripts, *tensors, **kwargs)
    elif len(tensors) <= 2:
        out = _contract(subscripts, *tensors, **kwargs)
    else:
        optimize = kwargs.pop('optimize', True)
        tensors = list(tensors)
        contraction_list = _einsum_path(subscripts, *tensors, optimize=optimize,
                                        einsum_call=True)[1]
        for contraction in contraction_list:
            inds, idx_rm, einsum_str, remaining = contraction[:4]
            tmp_operands = [tensors.pop(x) for x in inds]
            if len(tmp_operands) > 2:
                out = _numpy_einsum(einsum_str, *tmp_operands)
            else:
                out = contract(einsum_str, *tmp_operands)
            tensors.append(out)
    return out


# 2d -> 1d or 3d -> 2d
def pack_tril(mat, axis=-1, out=None):
    '''flatten the lower triangular part of a matrix.
    Given mat, it returns mat[...,numpy.tril_indices(mat.shape[0])]

    Examples:

    >>> pack_tril(numpy.arange(9).reshape(3,3))
    [0 3 4 6 7 8]
    '''
    if mat.size == 0:
        return numpy.zeros(mat.shape+(0,), dtype=mat.dtype)

    if mat.ndim == 2:
        count, nd = 1, mat.shape[0]
        shape = nd*(nd+1)//2
    else:
        count, nd = mat.shape[:2]
        shape = (count, nd*(nd+1)//2)

    if mat.ndim == 2 or axis == -1:
        mat = numpy.asarray(mat, order='C')
        out = numpy.ndarray(shape, mat.dtype, buffer=out)
        if mat.dtype == numpy.double:
            fn = _np_helper.NPdpack_tril_2d
        elif mat.dtype == numpy.complex128:
            fn = _np_helper.NPzpack_tril_2d
        else:
            out[:] = mat[numpy.tril_indices(nd)]
            return out

        fn(ctypes.c_int(count), ctypes.c_int(nd),
           out.ctypes.data_as(ctypes.c_void_p),
           mat.ctypes.data_as(ctypes.c_void_p))
        return out

    else:  # pack the leading two dimension
        assert (axis == 0)
        out = mat[numpy.tril_indices(nd)]
        return out

# 1d -> 2d or 2d -> 3d, write hermitian lower triangle to upper triangle
def unpack_tril(tril, filltriu=HERMITIAN, axis=-1, out=None):
    '''Reversed operation of pack_tril.

    Kwargs:
        filltriu : int

            | 0           Do not fill the upper triangular part, random number may appear
                          in the upper triangular part
            | 1 (default) Transpose the lower triangular part to fill the upper triangular part
            | 2           Similar to filltriu=1, negative of the lower triangular part is assign
                          to the upper triangular part to make the matrix anti-hermitian

    Examples:

    >>> unpack_tril(numpy.arange(6.))
    [[ 0. 1. 3.]
     [ 1. 2. 4.]
     [ 3. 4. 5.]]
    >>> unpack_tril(numpy.arange(6.), 0)
    [[ 0. 0. 0.]
     [ 1. 2. 0.]
     [ 3. 4. 5.]]
    >>> unpack_tril(numpy.arange(6.), 2)
    [[ 0. -1. -3.]
     [ 1.  2. -4.]
     [ 3.  4.  5.]]
    '''
    tril = numpy.asarray(tril, order='C')
    if tril.ndim == 1:
        count, nd = 1, tril.size
        nd = int(numpy.sqrt(nd*2))
        shape = (nd,nd)
    elif tril.ndim == 2:
        if axis == 0:
            nd, count = tril.shape
        else:
            count, nd = tril.shape
        nd = int(numpy.sqrt(nd*2))
        shape = (count,nd,nd)
    else:
        raise NotImplementedError('unpack_tril for high dimension arrays')

    if (tril.dtype != numpy.double and tril.dtype != numpy.complex128):
        out = numpy.ndarray(shape, tril.dtype, buffer=out)
        idx, idy = numpy.tril_indices(nd)
        if filltriu == ANTIHERMI:
            out[...,idy,idx] = -tril
        else:
            out[...,idy,idx] = tril
        out[...,idx,idy] = tril
        return out

    elif tril.ndim == 1 or axis == -1 or axis == tril.ndim-1:
        out = numpy.ndarray(shape, tril.dtype, buffer=out)
        if tril.dtype == numpy.double:
            fn = _np_helper.NPdunpack_tril_2d
        else:
            fn = _np_helper.NPzunpack_tril_2d
        fn(ctypes.c_int(count), ctypes.c_int(nd),
           tril.ctypes.data_as(ctypes.c_void_p),
           out.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(filltriu))
        return out

    else:  # unpack the leading dimension
        assert (axis == 0)
        shape = (nd,nd) + tril.shape[1:]
        out = numpy.ndarray(shape, tril.dtype, buffer=out)
        idx = numpy.tril_indices(nd)
        if filltriu == HERMITIAN:
            for ij,(i,j) in enumerate(zip(*idx)):
                out[i,j] = tril[ij]
                out[j,i] = tril[ij].conj()
        elif filltriu == ANTIHERMI:
            for ij,(i,j) in enumerate(zip(*idx)):
                out[i,j] = tril[ij]
                out[j,i] =-tril[ij].conj()
        elif filltriu == SYMMETRIC:
            #:for ij,(i,j) in enumerate(zip(*idx)):
            #:    out[i,j] = out[j,i] = tril[ij]
            idxy = numpy.empty((nd,nd), dtype=int)
            idxy[idx[0],idx[1]] = idxy[idx[1],idx[0]] = numpy.arange(nd*(nd+1)//2)
            numpy.take(tril, idxy, axis=0, out=out)
        else:
            out[idx] = tril
        return out

# extract a row from a tril-packed matrix
def unpack_row(tril, row_id):
    '''Extract one row of the lower triangular part of a matrix.
    It is equivalent to unpack_tril(a)[row_id]

    Examples:

    >>> unpack_row(numpy.arange(6.), 0)
    [ 0. 1. 3.]
    >>> unpack_tril(numpy.arange(6.))[0]
    [ 0. 1. 3.]
    '''
    tril = numpy.ascontiguousarray(tril)
    nd = int(numpy.sqrt(tril.size*2))
    mat = numpy.empty(nd, tril.dtype)
    if tril.dtype == numpy.double:
        fn = _np_helper.NPdunpack_row
    elif tril.dtype == numpy.complex128:
        fn = _np_helper.NPzunpack_row
    else:
        p0 = row_id*(row_id+1)//2
        p1 = row_id*(row_id+1)//2 + row_id
        idx = numpy.arange(row_id, nd)
        return numpy.append(tril[p0:p1], tril[idx*(idx+1)//2+row_id])

    fn.restype = ctypes.c_void_p
    fn(ctypes.c_int(nd), ctypes.c_int(row_id),
       tril.ctypes.data_as(ctypes.c_void_p),
       mat.ctypes.data_as(ctypes.c_void_p))
    return mat

# for i > j of 2d mat, mat[j,i] = mat[i,j]
def hermi_triu(mat, hermi=HERMITIAN, inplace=True):
    '''Use the elements of the lower triangular part to fill the upper triangular part.

    Kwargs:
        hermi : int

            | 1 (default) return a hermitian matrix
            | 2           return an anti-hermitian matrix

    Examples:

    >>> hermi_triu(numpy.arange(9.).reshape(3,3), 1)
    [[ 0.  3.  6.]
     [ 3.  4.  7.]
     [ 6.  7.  8.]]
    >>> hermi_triu(numpy.arange(9.).reshape(3,3), 2)
    [[ 0. -3. -6.]
     [ 3.  4. -7.]
     [ 6.  7.  8.]]
    '''
    assert (hermi == HERMITIAN or hermi == ANTIHERMI)
    if not inplace:
        mat = mat.copy('A')
    if mat.flags.c_contiguous:
        buf = mat
    elif mat.flags.f_contiguous:
        buf = mat.T
    else:
        raise NotImplementedError

    nd = mat.shape[0]
    assert (mat.size == nd**2)

    if mat.dtype == numpy.double:
        fn = _np_helper.NPdsymm_triu
    elif mat.dtype == numpy.complex128:
        fn = _np_helper.NPzhermi_triu
    else:
        raise NotImplementedError
    fn.restype = ctypes.c_void_p
    fn(ctypes.c_int(nd), buf.ctypes.data_as(ctypes.c_void_p),
       ctypes.c_int(hermi))
    return mat


LINEAR_DEP_THRESHOLD = 1e-10
def solve_lineq_by_SVD(a, b):
    '''Solving a * x = b.  If a is a singular matrix, its small SVD values are
    neglected.
    '''
    t, w, vH = numpy.linalg.svd(a)
    idx = []
    for i,wi in enumerate(w):
        if wi > LINEAR_DEP_THRESHOLD:
            idx.append(i)
    if idx:
        idx = numpy.array(idx)
        tb = numpy.dot(numpy.array(t[:,idx]).T.conj(), numpy.array(b))
        x = numpy.dot(numpy.array(vH[idx,:]).T.conj(), tb / w[idx])
    else:
        x = numpy.zeros_like(b)
    return x

def take_2d(a, idx, idy, out=None):
    '''Equivalent to a[idx[:,None],idy] for a 2D array.

    Examples:

    >>> out = numpy.arange(9.).reshape(3,3)
    >>> take_2d(a, [0,2], [0,2])
    [[ 0.  2.]
     [ 6.  8.]]
    '''
    a = numpy.asarray(a, order='C')
    out = numpy.ndarray((len(idx),len(idy)), dtype=a.dtype, buffer=out)
    idx = numpy.asarray(idx, dtype=numpy.int32)
    idy = numpy.asarray(idy, dtype=numpy.int32)
    if a.dtype == numpy.double:
        fn = _np_helper.NPdtake_2d
    elif a.dtype == numpy.complex128:
        fn = _np_helper.NPztake_2d
    else:
        return a[idx[:,None],idy]
    fn(out.ctypes.data_as(ctypes.c_void_p),
       a.ctypes.data_as(ctypes.c_void_p),
       idx.ctypes.data_as(ctypes.c_void_p),
       idy.ctypes.data_as(ctypes.c_void_p),
       ctypes.c_int(out.shape[1]), ctypes.c_int(a.shape[1]),
       ctypes.c_int(idx.size), ctypes.c_int(idy.size))
    return out

def takebak_2d(out, a, idx, idy, thread_safe=True):
    '''Reverse operation of take_2d.  Equivalent to out[idx[:,None],idy] += a
    for a 2D array.

    Examples:

    >>> out = numpy.zeros((3,3))
    >>> takebak_2d(out, numpy.ones((2,2)), [0,2], [0,2])
    [[ 1.  0.  1.]
     [ 0.  0.  0.]
     [ 1.  0.  1.]]
    '''
    assert (out.flags.c_contiguous)
    a = numpy.asarray(a, order='C')
    if out.dtype != a.dtype:
        a = a.astype(out.dtype)
    if out.dtype == numpy.double:
        fn = _np_helper.NPdtakebak_2d
    elif out.dtype == numpy.complex128:
        fn = _np_helper.NPztakebak_2d
    else:
        if thread_safe:
            out[idx[:,None], idy] += a
        else:
            raise NotImplementedError
        return out
    idx = numpy.asarray(idx, dtype=numpy.int32)
    idy = numpy.asarray(idy, dtype=numpy.int32)
    fn(out.ctypes.data_as(ctypes.c_void_p),
       a.ctypes.data_as(ctypes.c_void_p),
       idx.ctypes.data_as(ctypes.c_void_p),
       idy.ctypes.data_as(ctypes.c_void_p),
       ctypes.c_int(out.shape[1]), ctypes.c_int(a.shape[1]),
       ctypes.c_int(idx.size), ctypes.c_int(idy.size),
       ctypes.c_int(thread_safe))
    return out


def inplace_transpose_scale(a, alpha=1.0):
    """In-place parallel scaling and transposition of a square matrix

    Parameters
    ----------
    a : ndarray
        Square matrix of size (n,n) to be scaled and transposed.
        Does not need to be contiguous; lda can exceed n.
    alpha : float, optional
        scaling factor, by default 1.0
    """
    lda, order, _ = leading_dimension_order(a)
    assert a.shape[0] == a.shape[1]
    n = a.shape[0]
    assert order in ('C', 'F')
    if a.dtype == numpy.double:
        _np_helper.NPomp_d_itranspose_scale(
            ctypes.c_int(n), ctypes.c_double(alpha),
            a.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(lda)
        )
    elif a.dtype == numpy.complex128:
        alpha_arr = numpy.array([alpha], dtype=numpy.complex128)
        _np_helper.NPomp_z_itranspose_scale(
            ctypes.c_int(n), alpha_arr.ctypes.data_as(ctypes.c_void_p),
            a.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(lda)
        )
    else:
        raise NotImplementedError
    return a

def transpose(a, axes=None, inplace=False, out=None):
    '''Transposing an array with better memory efficiency

    Examples:

    >>> transpose(numpy.ones((3,2)))
    [[ 1.  1.  1.]
     [ 1.  1.  1.]]
    '''
    if inplace:
        if a.ndim == 2:
            inplace_transpose_scale(a)
        elif a.ndim == 3 and axes == (0,2,1):
            assert a.shape[1] == a.shape[2]
            astrides = [a.strides[i]//a.itemsize for i in (1, 2)]
            lda = max(astrides)
            assert min(astrides) == 1
            if a.dtype == numpy.double:
                _np_helper.NPomp_dtensor_itranspose_scale021(
                    ctypes.c_longlong(a.strides[0]//a.itemsize),
                    ctypes.c_int(a.shape[0]),
                    ctypes.c_int(a.shape[1]),
                    ctypes.c_double(1.0),
                    a.ctypes.data_as(ctypes.c_void_p),
                    ctypes.c_int(lda)
                )
            elif a.dtype == numpy.complex128:
                one_cplx = numpy.array([1.0], dtype=numpy.complex128)
                _np_helper.NPomp_ztensor_itranspose_scale021(
                    ctypes.c_longlong(a.strides[0]//a.itemsize),
                    ctypes.c_int(a.shape[0]),
                    ctypes.c_int(a.shape[1]),
                    one_cplx.ctypes.data_as(ctypes.c_void_p),
                    a.ctypes.data_as(ctypes.c_void_p),
                    ctypes.c_int(lda)
                )
            else:
                raise NotImplementedError
        else:
            arow, acol = a.shape[:2]
            assert arow == acol
            tmp = numpy.empty((BLOCK_DIM,BLOCK_DIM))
            for c0, c1 in misc.prange(0, acol, BLOCK_DIM):
                for r0, r1 in misc.prange(0, c0, BLOCK_DIM):
                    tmp[:c1-c0,:r1-r0] = a[c0:c1,r0:r1]
                    a[c0:c1,r0:r1] = a[r0:r1,c0:c1].T
                    a[r0:r1,c0:c1] = tmp[:c1-c0,:r1-r0].T
                # diagonal blocks
                a[c0:c1,c0:c1] = a[c0:c1,c0:c1].T
        return a

    if (not a.flags.c_contiguous
        or (a.dtype != numpy.double and a.dtype != numpy.complex128)):
        if a.ndim == 2:
            if a.flags.f_contiguous:
                return a.T.copy()

            arow, acol = a.shape
            out = numpy.empty((acol,arow), a.dtype)
            r1 = c1 = 0
            for c0 in range(0, acol-BLOCK_DIM, BLOCK_DIM):
                c1 = c0 + BLOCK_DIM
                for r0 in range(0, arow-BLOCK_DIM, BLOCK_DIM):
                    r1 = r0 + BLOCK_DIM
                    out[c0:c1,r0:r1] = a[r0:r1,c0:c1].T
                out[c0:c1,r1:arow] = a[r1:arow,c0:c1].T
            for r0 in range(0, arow-BLOCK_DIM, BLOCK_DIM):
                r1 = r0 + BLOCK_DIM
                out[c1:acol,r0:r1] = a[r0:r1,c1:acol].T
            out[c1:acol,r1:arow] = a[r1:arow,c1:acol].T
            return out
        else:
            return numpy.asarray(a.transpose(axes), order='C')

    if a.ndim == 2:
        arow, acol = a.shape
        c_shape = (ctypes.c_int*3)(1, arow, acol)
        out = numpy.ndarray((acol, arow), a.dtype, buffer=out)
    elif a.ndim == 3 and axes == (0,2,1):
        d0, arow, acol = a.shape
        c_shape = (ctypes.c_int*3)(d0, arow, acol)
        out = numpy.ndarray((d0, acol, arow), a.dtype, buffer=out)
    else:
        raise NotImplementedError

    assert (a.flags.c_contiguous)
    if a.dtype == numpy.double:
        fn = _np_helper.NPdtranspose_021
    else:
        fn = _np_helper.NPztranspose_021
    fn.restype = ctypes.c_void_p
    fn(c_shape, a.ctypes.data_as(ctypes.c_void_p),
       out.ctypes.data_as(ctypes.c_void_p))
    return out

def transpose_sum(a, inplace=False, out=None):
    '''Computing a + a.T with better memory efficiency

    Examples:

    >>> transpose_sum(numpy.arange(4.).reshape(2,2))
    [[ 0.  3.]
     [ 3.  6.]]
    '''
    return hermi_sum(a, inplace=inplace, out=out)

def hermi_sum(a, axes=None, hermi=HERMITIAN, inplace=False, out=None):
    '''Computing a + a.T.conj() with better memory efficiency

    Examples:

    >>> transpose_sum(numpy.arange(4.).reshape(2,2))
    [[ 0.  3.]
     [ 3.  6.]]
    '''
    if inplace:
        assert isinstance(a, numpy.ndarray)
        out = a
    else:
        a = numpy.asarray(a)
        out = numpy.ndarray(a.shape, a.dtype, buffer=out)

    if (not a.flags.c_contiguous
        or (a.dtype != numpy.double and a.dtype != numpy.complex128)):
        def _hermi_sum_2d(a, out):
            na = a.shape[0]
            for c0, c1 in misc.prange(0, na, BLOCK_DIM):
                for r0, r1 in misc.prange(0, c0, BLOCK_DIM):
                    tmp = a[r0:r1,c0:c1] + a[c0:c1,r0:r1].conj().T
                    out[c0:c1,r0:r1] = tmp.T.conj()
                    out[r0:r1,c0:c1] = tmp
                # diagonal blocks
                tmp = a[c0:c1,c0:c1] + a[c0:c1,c0:c1].conj().T
                out[c0:c1,c0:c1] = tmp
            return out

        if a.ndim == 2:
            return _hermi_sum_2d(a, out)
        elif a.ndim == 3 and axes == (0,2,1):
            assert a.shape[1] == a.shape[2]
            for i in range(a.shape[0]):
                _hermi_sum_2d(a[i], out[i])
            return out
        else:
            raise NotImplementedError('input array is not C-contiguous')

    if a.ndim == 2:
        assert (a.shape[0] == a.shape[1])
        c_shape = (ctypes.c_int*3)(1, a.shape[0], a.shape[1])
    elif a.ndim == 3 and axes == (0,2,1):
        assert (a.shape[1] == a.shape[2])
        c_shape = (ctypes.c_int*3)(*(a.shape))
    else:
        raise NotImplementedError

    assert (a.flags.c_contiguous)
    if a.dtype == numpy.double:
        fn = _np_helper.NPdsymm_021_sum
    else:
        fn = _np_helper.NPzhermi_021_sum
    fn(c_shape, a.ctypes.data_as(ctypes.c_void_p),
       out.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(hermi))
    return out

# NOTE: NOT assume array a, b to be C-contiguous, since a and b are two
# pointers we want to pass in.
# numpy.dot might not call optimized blas
def ddot(a, b, alpha=1, c=None, beta=0):
    '''Matrix-matrix multiplication for double precision arrays
    '''
    m = a.shape[0]
    k = a.shape[1]
    n = b.shape[1]
    if a.flags.c_contiguous:
        trans_a = 'N'
    elif a.flags.f_contiguous:
        trans_a = 'T'
        a = a.T
    else:
        a = numpy.asarray(a, order='C')
        trans_a = 'N'
        #raise ValueError('a.flags: %s' % str(a.flags))

    assert (k == b.shape[0])
    if b.flags.c_contiguous:
        trans_b = 'N'
    elif b.flags.f_contiguous:
        trans_b = 'T'
        b = b.T
    else:
        b = numpy.asarray(b, order='C')
        trans_b = 'N'
        #raise ValueError('b.flags: %s' % str(b.flags))

    if c is None:
        c = numpy.empty((m,n))
        beta = 0
    else:
        assert (c.shape == (m,n))

    return _dgemm(trans_a, trans_b, m, n, k, a, b, c, alpha, beta)

def zdot(a, b, alpha=1, c=None, beta=0):
    '''Matrix-matrix multiplication for double complex arrays
    '''
    m = a.shape[0]
    k = a.shape[1]
    n = b.shape[1]
    if a.flags.c_contiguous:
        trans_a = 'N'
    elif a.flags.f_contiguous:
        trans_a = 'T'
        a = a.T
    else:
        raise ValueError('a.flags: %s' % str(a.flags))

    assert (k == b.shape[0])
    if b.flags.c_contiguous:
        trans_b = 'N'
    elif b.flags.f_contiguous:
        trans_b = 'T'
        b = b.T
    else:
        raise ValueError('b.flags: %s' % str(b.flags))

    if c is None:
        beta = 0
        c = numpy.empty((m,n), dtype=numpy.complex128)
    else:
        assert (c.shape == (m,n))

    return _zgemm(trans_a, trans_b, m, n, k, a, b, c, alpha, beta)

def zdotNN(aR, aI, bR, bI, alpha=1, cR=None, cI=None, beta=0):
    '''c = a*b'''
    cR = ddot(aR, bR, alpha, cR, beta)
    cR = ddot(aI, bI,-alpha, cR, 1   )
    cI = ddot(aR, bI, alpha, cI, beta)
    cI = ddot(aI, bR, alpha, cI, 1   )
    return cR, cI

def zdotCN(aR, aI, bR, bI, alpha=1, cR=None, cI=None, beta=0):
    '''c = a.conj()*b'''
    cR = ddot(aR, bR, alpha, cR, beta)
    cR = ddot(aI, bI, alpha, cR, 1   )
    cI = ddot(aR, bI, alpha, cI, beta)
    cI = ddot(aI, bR,-alpha, cI, 1   )
    return cR, cI

def zdotNC(aR, aI, bR, bI, alpha=1, cR=None, cI=None, beta=0):
    '''c = a*b.conj()'''
    cR = ddot(aR, bR, alpha, cR, beta)
    cR = ddot(aI, bI, alpha, cR, 1   )
    cI = ddot(aR, bI,-alpha, cI, beta)
    cI = ddot(aI, bR, alpha, cI, 1   )
    return cR, cI

def dot(a, b, alpha=1, c=None, beta=0):
    atype = a.dtype
    btype = b.dtype

    if atype == numpy.float64 and btype == numpy.float64:
        if c is None or c.dtype == numpy.float64:
            return ddot(a, b, alpha, c, beta)
        else:
            cr = numpy.asarray(c.real, order='C')
            c.real = ddot(a, b, alpha, cr, beta)
            return c

    elif atype == numpy.complex128 and btype == numpy.complex128:
        # Gauss's complex multiplication algorithm may affect numerical stability
        #k1 = ddot(a.real+a.imag, b.real.copy(), alpha)
        #k2 = ddot(a.real.copy(), b.imag-b.real, alpha)
        #k3 = ddot(a.imag.copy(), b.real+b.imag, alpha)
        #ab = k1-k3 + (k1+k2)*1j
        return zdot(a, b, alpha, c, beta)

    elif atype == numpy.float64 and btype == numpy.complex128:
        if b.flags.f_contiguous:
            order = 'F'
        else:
            order = 'C'
        cr = ddot(a, numpy.asarray(b.real, order=order), alpha)
        ci = ddot(a, numpy.asarray(b.imag, order=order), alpha)
        ab = numpy.ndarray(cr.shape, dtype=numpy.complex128, buffer=c)
        if c is None or beta == 0:
            ab.real = cr
            ab.imag = ci
        else:
            ab *= beta
            ab.real += cr
            ab.imag += ci
        return ab

    elif atype == numpy.complex128 and btype == numpy.float64:
        if a.flags.f_contiguous:
            order = 'F'
        else:
            order = 'C'
        cr = ddot(numpy.asarray(a.real, order=order), b, alpha)
        ci = ddot(numpy.asarray(a.imag, order=order), b, alpha)
        ab = numpy.ndarray(cr.shape, dtype=numpy.complex128, buffer=c)
        if c is None or beta == 0:
            ab.real = cr
            ab.imag = ci
        else:
            ab *= beta
            ab.real += cr
            ab.imag += ci
        return ab

    else:
        if c is None:
            c = numpy.dot(a, b) * alpha
        elif beta == 0:
            c[:] = numpy.dot(a, b) * alpha
        else:
            c *= beta
            c += numpy.dot(a, b) * alpha
        return c

# a, b, c in C-order
def _dgemm(trans_a, trans_b, m, n, k, a, b, c, alpha=1, beta=0,
           offseta=0, offsetb=0, offsetc=0):
    if a.size == 0 or b.size == 0:
        if beta == 0:
            c[:] = 0
        else:
            c[:] *= beta
        return c

    assert (a.flags.c_contiguous)
    assert (b.flags.c_contiguous)
    assert (c.flags.c_contiguous)

    _np_helper.NPdgemm(ctypes.c_char(trans_b.encode('ascii')),
                       ctypes.c_char(trans_a.encode('ascii')),
                       ctypes.c_int(n), ctypes.c_int(m), ctypes.c_int(k),
                       ctypes.c_int(b.shape[1]), ctypes.c_int(a.shape[1]),
                       ctypes.c_int(c.shape[1]),
                       ctypes.c_int(offsetb), ctypes.c_int(offseta),
                       ctypes.c_int(offsetc),
                       b.ctypes.data_as(ctypes.c_void_p),
                       a.ctypes.data_as(ctypes.c_void_p),
                       c.ctypes.data_as(ctypes.c_void_p),
                       ctypes.c_double(alpha), ctypes.c_double(beta))
    return c
def _zgemm(trans_a, trans_b, m, n, k, a, b, c, alpha=1, beta=0,
           offseta=0, offsetb=0, offsetc=0):
    if a.size == 0 or b.size == 0:
        if beta == 0:
            c[:] = 0
        else:
            c[:] *= beta
        return c

    assert (a.flags.c_contiguous)
    assert (b.flags.c_contiguous)
    assert (c.flags.c_contiguous)
    assert (a.dtype == numpy.complex128)
    assert (b.dtype == numpy.complex128)
    assert (c.dtype == numpy.complex128)

    _np_helper.NPzgemm(ctypes.c_char(trans_b.encode('ascii')),
                       ctypes.c_char(trans_a.encode('ascii')),
                       ctypes.c_int(n), ctypes.c_int(m), ctypes.c_int(k),
                       ctypes.c_int(b.shape[1]), ctypes.c_int(a.shape[1]),
                       ctypes.c_int(c.shape[1]),
                       ctypes.c_int(offsetb), ctypes.c_int(offseta),
                       ctypes.c_int(offsetc),
                       b.ctypes.data_as(ctypes.c_void_p),
                       a.ctypes.data_as(ctypes.c_void_p),
                       c.ctypes.data_as(ctypes.c_void_p),
                       (ctypes.c_double*2)(alpha.real, alpha.imag),
                       (ctypes.c_double*2)(beta.real, beta.imag))
    return c

def frompointer(pointer, count, dtype=float):
    '''Interpret a buffer that the pointer refers to as a 1-dimensional array.

    Args:
        pointer : int or ctypes pointer
            address of a buffer
        count : int
            Number of items to read.
        dtype : data-type, optional
            Data-type of the returned array; default: float.

    Examples:

    >>> s = numpy.ones(3, dtype=numpy.int32)
    >>> ptr = s.ctypes.data
    >>> frompointer(ptr, count=6, dtype=numpy.int16)
    [1, 0, 1, 0, 1, 0]
    '''
    dtype = numpy.dtype(dtype)
    count *= dtype.itemsize
    buf = (ctypes.c_char * count).from_address(pointer)
    a = numpy.ndarray(count, dtype=numpy.int8, buffer=buf)
    return a.view(dtype)

def leading_dimension_order(a):
    """Return the leading dimension and the order of a matrix.

    Parameters
    ----------
    a : ndarray
        2D array.

    Returns
    -------
    lda : int
        Leading dimension of the array -- the stride between rows or columns.
    order : str
        'F' for col major, 'C' for row major, 'G' for neither.
    a_cshape : tuple
        If a is row major, a.shape; if a is col major, a.T.shape; otherwise None.
    """
    assert a.ndim == 2
    astrides = [s//a.itemsize for s in a.strides]
    lda = max(astrides)
    if astrides[0] == 1:
        order = 'F'
        a_cshape = a.T.shape
    elif astrides[1] == 1:
        order = 'C'
        a_cshape = a.shape
    else:
        order = 'G'
        a_cshape = None
    return lda, order, a_cshape

norm = numpy.linalg.norm

def cond(x, p=None):
    '''Compute the condition number'''
    if isinstance(x, numpy.ndarray) and x.ndim == 2 or p is not None:
        return numpy.linalg.cond(x, p)
    else:
        return numpy.asarray([numpy.linalg.cond(xi) for xi in x])

def cartesian_prod(arrays, out=None):
    '''
    Generate a cartesian product of input arrays.
    http://stackoverflow.com/questions/1208118/using-numpy-to-build-an-array-of-all-combinations-of-two-arrays

    Args:
        arrays : list of array-like
            1-D arrays to form the cartesian product of.
        out : ndarray
            Array to place the cartesian product in.

    Returns:
        out : ndarray
            2-D array of shape (M, len(arrays)) containing cartesian products
            formed of input arrays.

    Examples:

    >>> cartesian_prod(([1, 2, 3], [4, 5], [6, 7]))
    array([[1, 4, 6],
           [1, 4, 7],
           [1, 5, 6],
           [1, 5, 7],
           [2, 4, 6],
           [2, 4, 7],
           [2, 5, 6],
           [2, 5, 7],
           [3, 4, 6],
           [3, 4, 7],
           [3, 5, 6],
           [3, 5, 7]])

    '''
    arrays = [numpy.asarray(x) for x in arrays]
    dtype = numpy.result_type(*arrays)
    nd = len(arrays)
    dims = [nd] + [len(x) for x in arrays]
    out = numpy.ndarray(dims, dtype, buffer=out)

    shape = [-1] + [1] * nd
    for i, arr in enumerate(arrays):
        out[i] = arr.reshape(shape[:nd-i])

    return out.reshape(nd,-1).T

def direct_sum(subscripts, *operands):
    '''Apply the summation over many operands with the einsum fashion.

    Examples:

    >>> a = numpy.random.random((6,5))
    >>> b = numpy.random.random((4,3,2))
    >>> direct_sum('ij,klm->ijklm', a, b).shape
    (6, 5, 4, 3, 2)
    >>> direct_sum('ij,klm', a, b).shape
    (6, 5, 4, 3, 2)
    >>> direct_sum('i,j,klm->mjlik', a[0], a[:,0], b).shape
    (2, 6, 3, 5, 4)
    >>> direct_sum('ij-klm->ijklm', a, b).shape
    (6, 5, 4, 3, 2)
    >>> direct_sum('ij+klm', a, b).shape
    (6, 5, 4, 3, 2)
    >>> direct_sum('-i-j+klm->mjlik', a[0], a[:,0], b).shape
    (2, 6, 3, 5, 4)
    >>> c = numpy.random((3,5))
    >>> z = direct_sum('ik+jk->kij', a, c).shape  # This is slow
    >>> abs(a.T.reshape(5,6,1) + c.reshape(5,1,3) - z).sum()
    0.0
    '''

    def sign_and_symbs(subscript):
        ''' sign list and notation list'''
        subscript = subscript.replace(' ', '').replace(',', '+')

        if subscript[0] not in '+-':
            subscript = '+' + subscript
        sign = [x for x in subscript if x in '+-']

        symbs = subscript[1:].replace('-', '+').split('+')
        #s = ''.join(symbs)
        #assert (len(set(s)) == len(s))  # make sure no duplicated symbols
        return sign, symbs

    if '->' in subscripts:
        src, dest = subscripts.split('->')
        sign, src = sign_and_symbs(src)
        dest = dest.replace(' ', '')
    else:
        sign, src = sign_and_symbs(subscripts)
        dest = ''.join(src)
    assert (len(src) == len(operands))

    for i, symb in enumerate(src):
        op = numpy.asarray(operands[i])
        assert (len(symb) == op.ndim)
        unisymb = set(symb)
        if len(unisymb) != len(symb):
            unisymb = ''.join(unisymb)
            op = _numpy_einsum('->'.join((symb, unisymb)), op)
            src[i] = unisymb
        if i == 0:
            if sign[i] == '+':
                out = op
            else:
                out = -op
        elif sign[i] == '+':
            out = out.reshape(out.shape+(1,)*op.ndim) + op
        else:
            out = out.reshape(out.shape+(1,)*op.ndim) - op

    out = _numpy_einsum('->'.join((''.join(src), dest)), out)
    out.flags.writeable = True  # old numpy has this issue
    return out

def condense(opname, a, loc_x, loc_y=None):
    '''
    .. code-block:: python

        for i,i0 in enumerate(loc_x[:-1]):
            i1 = loc_x[i+1]
            for j,j0 in enumerate(loc_y[:-1]):
                j1 = loc_y[j+1]
                out[i,j] = op(a[i0:i1, j0:j1])
    '''
    assert a.ndim == 2
    if loc_y is None:
        loc_y = loc_x
    loc_x = numpy.asarray(loc_x, numpy.int32)
    loc_y = numpy.asarray(loc_y, numpy.int32)
    nloc_x = loc_x.size - 1
    nloc_y = loc_y.size - 1
    opname = opname.replace('numpy.', '').replace('np.', '')
    if opname.startswith('NP_'):
        opname = opname[3:]

    if (a.dtype == numpy.double and
        opname in ('sum', 'max', 'min', 'abssum', 'absmax', 'absmin', 'norm')):
        op = getattr(_np_helper, 'NP_' + opname)
        if a.flags.f_contiguous:
            a = transpose(a.T)
        a = numpy.asarray(a, order='C')
        out = numpy.zeros((nloc_x, nloc_y))
        _np_helper.NPcondense(op, out.ctypes.data_as(ctypes.c_void_p),
                              a.ctypes.data_as(ctypes.c_void_p),
                              loc_x.ctypes.data_as(ctypes.c_void_p),
                              loc_y.ctypes.data_as(ctypes.c_void_p),
                              ctypes.c_int(nloc_x), ctypes.c_int(nloc_y))
        return out

    if a.dtype in (bool, numpy.int8) and opname in ('any', 'all'):
        op = getattr(_np_helper, 'NP_' + opname)
        if a.flags.f_contiguous:
            a = transpose(a.T)
        a = numpy.asarray(a, order='C')
        out = numpy.zeros((nloc_x, nloc_y), dtype=a.dtype)
        _np_helper.NPbcondense(op, out.ctypes.data_as(ctypes.c_void_p),
                               a.ctypes.data_as(ctypes.c_void_p),
                               loc_x.ctypes.data_as(ctypes.c_void_p),
                               loc_y.ctypes.data_as(ctypes.c_void_p),
                               ctypes.c_int(nloc_x), ctypes.c_int(nloc_y))
        return out

    tmp = numpy.empty((nloc_x, a.shape[1]), dtype=a.dtype)
    out = numpy.empty((nloc_x, nloc_y), dtype=a.dtype)
    op = getattr(numpy, opname)
    for i, (i0, i1) in enumerate(zip(loc_x[:-1], loc_x[1:])):
        tmp[i] = op(a[i0:i1], axis=0)
    for j, (j0, j1) in enumerate(zip(loc_y[:-1], loc_y[1:])):
        out[:,j] = op(tmp[:,j0:j1], axis=1)
    return out

def expm(a):
    '''Equivalent to scipy.linalg.expm'''
    bs = [a.copy()]
    n = 0
    for n in range(1, 14):
        bs.append(ddot(bs[-1], a))
        radius = (2**(n*(n+2))*math.factorial(n+2)*1e-16) **((n+1.)/(n+2))
        #print(n, radius, bs[-1].max(), -bs[-1].min())
        if bs[-1].max() < radius and -bs[-1].min() < radius:
            break

    y = numpy.eye(a.shape[0])
    fac = 1
    for i, b in enumerate(bs):
        fac *= i + 1
        b *= (.5**(n*(i+1)) / fac)
        y += b
    buf, bs = bs[0], None
    for i in range(n):
        ddot(y, y, 1, buf, 0)
        y, buf = buf, y
    return y

def omatcopy(a, out=None):
    """Copies a matrix.

    Parameters
    ----------
    a : ndarray
        Matrix to be copied. The order of the matrix is preserved.
        a can be either row or column major.
    out : ndarray, optional
        Matrix to be overwritten. A new one is allocated if not provided.

    Returns
    -------
    out : ndarray
        Copy of a with the same order.
    """
    lda, _, a_cshape = leading_dimension_order(a)
    if out is None:
        out = numpy.empty_like(a)
    ld_out, _, out_cshape = leading_dimension_order(out)
    assert out_cshape == a_cshape and a_cshape is not None
    if a.dtype == numpy.double:
        fn = _np_helper.NPomp_dcopy
    elif a.dtype == numpy.complex128:
        fn = _np_helper.NPomp_zcopy
    else:
        raise NotImplementedError
    fn(ctypes.c_size_t(a_cshape[0]),
       ctypes.c_size_t(a_cshape[1]),
       a.ctypes.data_as(ctypes.c_void_p),
       ctypes.c_size_t(lda),
       out.ctypes.data_as(ctypes.c_void_p),
       ctypes.c_size_t(ld_out))
    return out

def zeros(shape, dtype=numpy.double, order='C'):
    """Allocate and zero an array in parallel. Useful for multi-socket systems
       due to the first touch policy. On most systems np.zeros does not count
       as first touch. Arrays returned by this function will (ideally) have
       pages backing them that are distributed across the sockets.
    """
    dtype = numpy.dtype(dtype)
    if dtype == numpy.double:
        out = numpy.empty(shape, dtype=dtype, order=order)
        _np_helper.NPomp_dset0(ctypes.c_size_t(out.size),
                              out.ctypes.data_as(ctypes.c_void_p))
    elif dtype == numpy.complex128:
        out = numpy.empty(shape, dtype=dtype, order=order)
        _np_helper.NPomp_zset0(ctypes.c_size_t(out.size),
                              out.ctypes.data_as(ctypes.c_void_p))
    else: # fallback
        out = numpy.zeros(shape, dtype=dtype, order=order)
    return out

def entrywise_mul(a, b, out=None):
    """Entrywise multiplication of two matrices.

    Parameters
    ----------
    a : ndarray
    b : ndarray
    out : ndarray, optional
        Output matrix. A new one is allocated if not provided.

    Returns
    -------
    ndarray
        a * b
    """
    assert a.ndim == 2 and b.ndim == 2
    assert a.shape == b.shape and a.dtype == b.dtype
    lda, _, a_cshape = leading_dimension_order(a)
    ldb, _, b_cshape = leading_dimension_order(b)
    if out is None:
        out = numpy.empty_like(b)
    ld_out, _, out_cshape = leading_dimension_order(out)
    assert a_cshape == b_cshape and b_cshape == out_cshape and a_cshape is not None
    if a.dtype == numpy.double:
        fn = _np_helper.NPomp_dmul
    elif a.dtype == numpy.complex128:
        fn = _np_helper.NPomp_zmul
    else:
        return numpy.multiply(a, b, out=out)
    fn(ctypes.c_size_t(a_cshape[0]),
       ctypes.c_size_t(a_cshape[1]),
       a.ctypes.data_as(ctypes.c_void_p),
       ctypes.c_size_t(lda),
       b.ctypes.data_as(ctypes.c_void_p),
       ctypes.c_size_t(ldb),
       out.ctypes.data_as(ctypes.c_void_p),
       ctypes.c_size_t(ld_out))
    return out

def ndarray_pointer_2d(array):
    '''Return an array that contains the addresses of the first element in each
    row of the input 2d array.
    '''
    assert array.ndim == 2
    assert array.flags.c_contiguous
    i = numpy.arange(array.shape[0])
    return array.ctypes.data + (i * array.strides[0]).astype(numpy.uintp)

class NPArrayWithTag(numpy.ndarray):
    # Initialize kwargs in function tag_array
    #def __new__(cls, a, **kwargs):
    #    obj = numpy.asarray(a).view(cls)
    #    obj.__dict__.update(kwargs)
    #    return obj

    # Customize __reduce__ and __setstate__ to keep tags after serialization
    # pickle.loads(pickle.dumps(tagarray)).  This is needed by mpi communication
    def __reduce__(self):
        pickled = numpy.ndarray.__reduce__(self)
        state = pickled[2] + (self.__dict__,)
        return (pickled[0], pickled[1], state)

    def __setstate__(self, state):
        numpy.ndarray.__setstate__(self, state[0:-1])
        self.__dict__.update(state[-1])

    # Whenever the contents of the array were modified (through ufunc), the tag
    # should be expired. Overwrite the output of ufunc to restore ndarray type.
    def __array_wrap__(self, out, context=None, return_scalar=False):
        if out.ndim == 0:  # if ufunc returns a scalar
            return out[()]
        else:
            return out.view(numpy.ndarray)


def tag_array(a, **kwargs):
    '''Attach attributes to numpy ndarray. The attribute name and value are
    obtained from the keyword arguments.
    '''
    # Make a shadow copy in any circumstance by converting it to an nparray.
    # If a is an object of NPArrayWithTag, all attributes will be lost in this
    # conversion. They need to be restored.
    t = numpy.asarray(a).view(NPArrayWithTag)

    if isinstance(a, NPArrayWithTag):
        t.__dict__.update(a.__dict__)
    t.__dict__.update(kwargs)
    return t

#TODO: merge with function pbc.cc.kccsd_rhf.vector_to_nested
def split_reshape(a, shapes):
    '''
    Split a vector into multiple tensors. shapes is a list of tuples.
    The entries of shapes indicate the shape of each tensor.

    Returns:
        tensors : a list of tensors

    Examples:

    >>> a = numpy.arange(12)
    >>> split_reshape(a, ((2,3), (1,), ((2,2), (1,1))))
    [array([[0, 1, 2],
            [3, 4, 5]]),
     array([6]),
     [array([[ 7,  8],
             [ 9, 10]]),
      array([[11]])]]
    '''
    if isinstance(shapes[0], (int, numpy.integer)):
        return a.reshape(shapes)

    def sub_split(a, shapes):
        tensors = []
        p1 = 0
        for shape in shapes:
            if isinstance(shape[0], (int, numpy.integer)):
                p0, p1 = p1, p1 + numpy.prod(shape)
                tensors.append(a[p0:p1].reshape(shape))
            else:
                subtensors, size = sub_split(a[p1:], shape)
                p1 += size
                tensors.append(subtensors)
        size = p1
        return tensors, size
    return sub_split(a, shapes)[0]

def locs_to_indices(locs, segement_list):
    '''
    Generate indices based on the segement information list "locs" and the
    required segements.

    Args:
        locs : list or ndarray
            locs[i], locs[i+1] indicates the [start:end] index for i-th segement
        segement_list: list or ndarray
            The segement Ids to extract.

    Examples::

    >>> locs_to_indices([0, 2, 5, 6, 9, 15, 17], [0, 2, 3, 5])
    array([0, 1, 5, 6, 7, 8, 15, 16])

    >>> locs_to_indices([0, 2, 5, 6, 9], array([True, False, True, True]))
    array([0, 1, 5, 6, 7, 8])
    '''
    segement_list = numpy.asarray(segement_list)
    if segement_list.dtype == bool:
        segement_list = numpy.where(segement_list)[0]
    nsegs = segement_list.size
    if nsegs == 0:
        idx = []
    elif nsegs == 1:
        i0, i1 = locs[segement_list[0]:segement_list[0]+2]
        idx = numpy.arange(i0, i1)
    else:
        assert segement_list[-1] < len(locs)
        # Find the dis-continued Ids in segement_list
        bas_mask = segement_list[:-1] + 1 != segement_list[1:]
        segs_loc = numpy.hstack([0, numpy.where(bas_mask)[0] + 1, nsegs])
        idx = []
        for b0, b1 in zip(segs_loc[:-1], segs_loc[1:]):
            i0, i1 = locs[segement_list[b0]], locs[segement_list[b1-1]+1]
            idx.append(numpy.arange(i0, i1))
        idx = numpy.hstack(idx)
    return numpy.asarray(idx, dtype=numpy.int32)

def cleanse(a, axis=0, tol=0):
    '''
    Remove floating-point errors by setting the
    numbers with differences smaller than `tol`
    to the same value. This should allow
    `numpy.round_` and `numpy.unique` together
    to work as expected.

    Args:
        a : ndarray
            Array to be cleansed.
        axis : int or None
            Axis along which the array values are compared.
            Default is the first axis. If set to None,
            the flattened array is used.
        tol : floating
            Tolerance, default is 0.
    Returns:
        Cleansed array.
    '''
    def _cleanse_1d(a_flat, tol):
        sorted_index = numpy.argsort(a_flat, axis=None)
        sorted_a_flat = a_flat[sorted_index]
        diff = numpy.diff(sorted_a_flat)
        cluster_loc = numpy.append(numpy.append(0, numpy.argwhere(diff > tol)[:,0]+1), a_flat.size)
        for i in range(len(cluster_loc)-1):
            id0, id1 = cluster_loc[i], cluster_loc[i+1]
            a_flat[sorted_index[id0:id1]] = a_flat[sorted_index[id0]]
        return a_flat

    if axis is None:
        a_flat = a.flatten()
        return _cleanse_1d(a_flat, tol).reshape(a.shape)
    else:
        a0 = numpy.moveaxis(a, axis, -1)
        shape = a0.shape
        a0 = a0.reshape(-1, a0.shape[-1])
        out = []
        for i in range(len(a0)):
            out.append(_cleanse_1d(a0[i].flatten(), tol))
        out = numpy.asarray(out).reshape(shape)
        return numpy.moveaxis(out, -1, axis)

def base_repr_int(number, base, ndigits=None):
    '''
    Similar to numpy.base_repr, but returns a list of integers.

    Args:
        number : array or int
            The value to convert. Negative values are converted to
            their absolute values.
        base : int
            Convert `number` to the `base` number system.
        ndigits : int, optional
            Number of digits. If given, pad zeros to the left until the number
            of digits reaches `ndigits`. Default is None, meaning no padding.

    Returns:
        out : list
            Representation of `number` in `base` system.

    Examples::

    >>> lib.base_repr_int(29, 8)
    [3, 5]

    >>> lib.base_repr_int(29, 8, 3)
    [0, 3, 5]
    '''
    if isinstance(number, numpy.ndarray):
        assert ndigits is not None
        number = number.flatten()
        res = numpy.empty([ndigits, len(number)], dtype=int)
        for i in range(ndigits-1, -1, -1):
            ki = number // base**i
            number -= ki * base**i
            res[ndigits-1-i] = ki
        return res.T

    num = abs(number)
    res = []
    if num == 0:
        res = [0]
    while num:
        res.append(num % base)
        num //= base
    if ndigits:
        padding = ndigits - len(res)
        res += [0] * padding
    res.reverse()
    return res

def inv_base_repr_int(x, base):
    '''Inverse of `base_repr_int`.
    Similar to Python function int(), but for arbitrary base.

    Args:
        x : array like
        base : int

    Returns:
        out : int

    Examples::

    >>> lib.inv_base_repr_int([0, 18, 9], 27)
    495

    >>> lib.base_repr_int(495, 27, 3)
    [0, 18, 9]
    '''
    out = 0
    x = numpy.asarray(x, dtype=int)
    if x.ndim > 1:
        shape = x.shape
        nd = shape[-1]
        x = x.reshape(-1, nd)
        for i in range(nd):
            out += x[:,i] * base ** (nd-i-1)
        out = out.reshape(shape[:-1])
    else:
        for i, ix in enumerate(x[::-1]):
            out += ix * base**i
    return out

def isin_1d(v, vs, return_index=False):
    '''Check if vector `v` is in vectors `vs`.

    Args:
        v : array like
            The target vector. `v` is flattened.
        vs : array like
            A list of vectors. The last dimension of `vs`
            should be the same as the size of `v`.
        return_index : bool
            Index of `v` in `vs`.

    Examples::

    >>> lib.isin_1d([1,2], [[2,1],[1,2]])
    True

    >>> lib.isin_1d([1,2], [[2,1],[2,1]])
    False
    '''
    v = numpy.asarray(v).flatten()
    n = len(v)
    vs = numpy.asarray(vs).reshape(-1, n)
    diff = abs(v[None,:] - vs)
    diff = numpy.sum(diff, axis=1)
    idx = numpy.where(diff == 0)[0]
    if len(idx) > 0:
        v_in_vs = True
    else:
        v_in_vs = False

    if not return_index:
        return v_in_vs
    else:
        if len(idx) == 1:
            idx = idx[0]
        return v_in_vs, idx

if __name__ == '__main__':
    a = numpy.random.random((30,40,5,10))
    b = numpy.random.random((10,30,5,20))
    c = numpy.random.random((10,20,20))
    d = numpy.random.random((20,10))
    f = einsum('ijkl,xiky,ayp,px->ajl', a,b,c,d, optimize=True)
    ref = einsum('ijkl,xiky->jlxy', a, b)
    ref = einsum('jlxy,ayp->jlxap', ref, c)
    ref = einsum('jlxap,px->ajl', ref, d)
    print(abs(ref-f).max())



'''
This lib loads results that are in a QCSchema format json.
Funcs below can recreate the mol and scf objects from the info in the json.
'''
import json
import numpy as np
import pyscf
from pyscf.lib.parameters import BOHR

def load_qcschema_json( file_name ):
    '''
    Does: loads qcschema format json into a dictionary
    Input:
        file_name: qcschema format json file

    Returns: dict in qcschema format
    '''
    # load qcschema output json file
    data = None
    with open(file_name,'r') as f:
        data = json.load(f)
    return data

def load_qcschema_go_final_json( file_name ):
    '''
    Does: loads qcschema format geometry optimization json
          and returns only the optimized 'final' geometry
          qcschema info as a dictionary.

    Input:
        file_name: qcschema format json file

    Returns: dict in qcschema format
    '''
    # load qcschema GO output json file
    # and return last 'trajectory' point's entries
    # (this is the optimized molecule)
    data = None
    temp = None
    with open(file_name,'r') as f:
        temp = json.load(f)
    data = temp["trajectory"][-1]
    return data

def load_qcschema_molecule(qcschema_dict, to_Angstrom=False, xyz=False, mol_select=1, step=0):
    '''
    Does: Loads molecule from qcschema format dict.
          Molecule may be single point molecule or from a geometry optimization/trajectory.

    Input:
        syms: atom symbols (qcschema format)
        coords: x y z coordinates (in qcschema format)
        mol_select: specifies which molecule to load from qcschema format results.
                    Default loads 'molecule' from qcschema.
                    Geometry optimizations or trajectories have mutliple geometries saved in the schema.
                    mol_select = 1 (default) molecule from standard qcschema format
                               = 2 initial_molecule in GO or traj qcschema
                               = 3 final_molecule in GO or traj qcschema
                               = 4 a specific step in the GO or traj qcschema, specify with 'step' arg.
        step: for geometry optimization or trajectory, which have multiple molecules in a qcschema output.
              This specifies which step to load the molecule from.
        to_Angstrom (optional): convert coordinates to Angstrom (default is Bohr)
        xyz (optional): controls output (see below)

    Returns:
        xyz=False (default): 'atom x y z' format string
        xyz=True: output a string in xyz file format
                  i.e. first line is number of atoms.
    '''
    if(mol_select == 1):
        syms = np.array(qcschema_dict["molecule"]["symbols"])
        geo = np.array(qcschema_dict["molecule"]["geometry"])
    elif(mol_select == 2):
        syms = np.array(qcschema_dict["initial_molecule"]["symbols"])
        geo = np.array(qcschema_dict["initial_molecule"]["geometry"])
    elif(mol_select == 3):
        syms = np.array(qcschema_dict["final_molecule"]["symbols"])
        geo = np.array(qcschema_dict["final_molecule"]["geometry"])
    elif(mol_select == 4):
        # for geometry or md, can load a specific geometry
        syms = np.array(qcschema_dict["trajectory"][step]["molecule"]["symbols"])
        geo = np.array(qcschema_dict["trajectory"][step]["molecule"]["geometry"])

    if(to_Angstrom):
        # convert Bohr to Angstrom
        geo = geo*BOHR

    NAtoms = len(syms)
    geo = np.reshape(geo, (NAtoms,3))

    PySCF_atoms = list(zip(syms, geo))

    # Return as string or return as xyz-format string (i.e. top is NAtoms,blankline)
    if(xyz):
        bldstr = f'{NAtoms}\n\n'
        for element, coordinates in PySCF_atoms:
            bldstr += f'{element} {coordinates[0]}, {coordinates[1]}, {coordinates[2]}\n'
            PySCF_atoms = bldstr
    return PySCF_atoms

def load_qcschema_hessian(qcschema_dict):
    '''
    Does: loads hessian from qcschema format dictionary
    Input:
        qcschema_dict

    Returns: hessian with format (N,N,3,3)
    '''
    # qcschema_dict: pass in dict that has the qcschema output json loaded into it

    # load qcschema hessian
    qc_h = []
    qc_h = qcschema_dict["return_result"]

    # Get Number of atoms N
    syms = np.array(qcschema_dict["molecule"]["symbols"])
    NAtom = len(syms)

    # reshape from (3N)**2 array to (N,N,3,3)
    hessian = np.array(qc_h).reshape(NAtom,NAtom,3,3)
    return hessian

def load_qcschema_scf_info(qcschema_dict):
    '''
    Does: loads scf info from qcschema format dictionary
    Input:
        qcschema_dict

    Returns:
        scf_dict: contains the relevent scf info only
    '''

    # Restricted wfn has schema scf_occupations_a occ of 1 or 0.
    # Need to double if rhf/rks/rohf
    method = qcschema_dict["keywords"]["scf"]["method"]
    if(method == 'rks' or method == 'roks' or method == 'rhf' or method == 'rohf'):
        OccFactor = 2.0
        have_beta = False
    elif(method == 'uks' or method == 'uhf'):
        OccFactor = 1.0
        have_beta = True
    elif(method == 'gks' or method == 'ghf'):
        OccFactor = 1.0
        have_beta = False
    else:
        raise RuntimeError('qcschema: cannot determine method..exit')
        return

    # need to reshape MO coefficients for PySCF shape.
    nao = qcschema_dict["properties"]["calcinfo_nbasis"]
    # nmo info often missing
    try:
        nmo = qcschema_dict["properties"]["calcinfo_nmo"]
    except KeyError:
        # key not provided..so we make an assumption
        # note: assumes nmo=nao which isn't the case if linear dependencies etc.
        # ..so may give error when reading coeffs
        nmo = nao
    assert nmo == nao

    # get the 4 things that PySCF wants
    # ...remembering to reshape coeffs and scale occupancies.
    e_tot = float( qcschema_dict["properties"]["return_energy"] )
    mo_coeff = np.reshape(qcschema_dict["wavefunction"]["scf_orbitals_a"],(nao,nmo))
    mo_occ = np.array( qcschema_dict["wavefunction"]["scf_occupations_a"] )*OccFactor
    mo_energy = np.array( qcschema_dict["wavefunction"]["scf_eigenvalues_a"] )
    if(have_beta):
        # for each useful piece of info we need to combine alpha and beta into 2d array, with alpha first
        # MO occupations
        mo_occ_beta = qcschema_dict["wavefunction"]["scf_occupations_b"]
        mo_occ = np.vstack( (mo_occ, mo_occ_beta) )
        # MO coefficients
        mo_coeff_beta = np.reshape(qcschema_dict["wavefunction"]["scf_orbitals_b"],(nao,nmo))
        mo_coeff = np.vstack( (mo_coeff,mo_coeff_beta))
        mo_coeff = np.reshape(mo_coeff,(2,nao,nmo))
        # MO energies
        mo_energy_beta = np.array( qcschema_dict["wavefunction"]["scf_eigenvalues_b"] )
        mo_energy = np.vstack( (mo_energy, mo_energy_beta) )
        # etot obviously doesn't need manipulation

    # convert to dictionary for PySCF
    scf_dic = {'e_tot'    : e_tot,
               'mo_energy': mo_energy,
               'mo_occ'   : mo_occ,
               'mo_coeff' : mo_coeff}

    return scf_dic


def recreate_mol_obj(qcschema_dict,to_Angstrom=False):
    '''
    Does: recreates mol object from qcschema format dictionary
    Input:
        qcschema_dict
        to_Angstrom: optional bool to convert geometry to Angstrom (default is Bohr)

    Returns: mol object
    '''

    ## Mol info: ##
    PySCF_charge = int( qcschema_dict["molecule"]["molecular_charge"] )
    # PySCF 'spin' is number of unpaired electrons, it will be mult-1
    PySCF_spin = int( qcschema_dict["molecule"]["molecular_multiplicity"] ) - 1
    PySCF_basis = str( qcschema_dict["model"]["basis"] )

    # Cartesian/Pure basis
    PySCF_cart = bool( qcschema_dict["keywords"]["basisSet"]["cartesian"] )

    # Get molecular structure.
    PySCF_atoms = load_qcschema_molecule(qcschema_dict, to_Angstrom,False)

    # Unit Bohr or Angstrom. QCSchema default is Bohr but can change here.
    if(to_Angstrom):
        units='A'
    else:
        units='B'

    ## Create mol ##
    mol = pyscf.gto.Mole(atom=PySCF_atoms,basis=PySCF_basis,ecp=PySCF_basis,
                         charge=PySCF_charge,spin=PySCF_spin,cart=PySCF_cart,unit=units)
    mol.build(False,False)

    return mol

def recreate_scf_obj(qcschema_dict,mol):
    '''
    Does: recreates scf object from qcschema format dictionary
    Input:
        qcschema_dict
        mol object

    Returns: scf object
    '''
    # load info from qcschema needed for scf obj
    scf_dict = load_qcschema_scf_info(qcschema_dict)

    # create scf object
    method =  qcschema_dict["keywords"]["scf"]["method"]
    if(method =='rks'):
        ks = mol.RKS()
    elif(method =='uks'):
        ks = mol.UKS()
    elif(method =='rhf'):
        ks = mol.RHF()
    elif(method =='uhf'):
        ks = mol.UHF()
    elif(method =='gks'):
        ks = mol.GKS()
    elif(method =='ghf'):
        ks = mol.GHF()
    else:
        raise RuntimeError('qcschema: cannot determine method..exit')
        return

    # get functional
    if(method == 'rks' or method == 'uks' or method == 'gks'):
        functional = qcschema_dict["keywords"]["xcFunctional"]["name"]
        ks.xc = functional

    # Load 4 key pieces of info we got from json into SCF object
    ks.mo_coeff = scf_dict["mo_coeff"]
    ks.mo_energy = scf_dict["mo_energy"]
    ks.mo_occ = scf_dict["mo_occ"]
    ks.e_tot = scf_dict["e_tot"]
    return ks



#!/usr/bin/env python
# Copyright 2014-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Interface to geomeTRIC library https://github.com/leeping/geomeTRIC
'''

import os
import tempfile
import numpy
import geometric
import geometric.molecule
from pyscf import lib
from pyscf.geomopt.addons import dump_mol_geometry
from pyscf import __config__
from pyscf.pbc.grad.krhf import GradientsBase

try:
    from geometric import internal, optimize, nifty, engine, molecule
except ImportError:
    msg = ('Geometry optimizer geomeTRIC not found.\ngeomeTRIC library '
           'can be found on github https://github.com/leeping/geomeTRIC.\n'
           'You can install geomeTRIC with "pip install geometric"')
    raise ImportError(msg)

# Overwrite units defined in geomeTRIC
internal.ang2bohr = optimize.ang2bohr = nifty.ang2bohr = 1./lib.param.BOHR
engine.bohr2ang = internal.bohr2ang = molecule.bohr2ang = nifty.bohr2ang = \
        optimize.bohr2ang = lib.param.BOHR
del (internal, optimize, nifty, engine, molecule)


INCLUDE_GHOST = getattr(__config__, 'geomopt_berny_solver_optimize_include_ghost', True)
ASSERT_CONV = getattr(__config__, 'geomopt_berny_solver_optimize_assert_convergence', True)

class PySCFEngine(geometric.engine.Engine):
    def __init__(self, scanner):
        molecule = geometric.molecule.Molecule()
        self.cell = cell = scanner.cell
        molecule.elem = [cell.atom_symbol(i) for i in range(cell.natm)]
        # Molecule is the geometry parser for a bunch of formats which use
        # Angstrom for Cartesian coordinates by default.
        molecule.xyzs = [cell.atom_coords()*lib.param.BOHR]  # In Angstrom
        super().__init__(molecule)

        self.scanner = scanner
        self.cycle = 0
        self.e_last = 0
        self.callback = None
        self.maxsteps = 100
        self.assert_convergence = False

    def calc_new(self, coords, dirname):
        if self.cycle >= self.maxsteps:
            raise NotConvergedError('Geometry optimization is not converged in '
                                    '%d iterations' % self.maxsteps)

        g_scanner = self.scanner
        cell = self.cell
        self.cycle += 1
        lib.logger.note(g_scanner, '\nGeometry optimization cycle %d', self.cycle)

        # geomeTRIC requires coords and gradients in atomic unit
        coords = coords.reshape(-1,3)
        if g_scanner.verbose >= lib.logger.NOTE:
            dump_mol_geometry(cell, coords*lib.param.BOHR)

        cell.set_geom_(coords, unit='Bohr')
        energy, gradients = g_scanner(cell)
        lib.logger.note(g_scanner,
                        'cycle %d: E = %.12g  dE = %g  norm(grad) = %g', self.cycle,
                        energy, energy - self.e_last, numpy.linalg.norm(gradients))
        self.e_last = energy

        if callable(self.callback):
            self.callback(locals())

        if self.assert_convergence and not g_scanner.converged:
            raise RuntimeError('Nuclear gradients of %s not converged' % g_scanner.base)
        return {"energy": energy, "gradient": gradients.ravel()}

def kernel(method, assert_convergence=ASSERT_CONV,
           include_ghost=INCLUDE_GHOST, constraints=None, callback=None,
           maxsteps=100, **kwargs):
    '''Optimize geometry with geomeTRIC library for the given method.

    To adjust the convergence threshold, parameters can be set in kwargs as
    below:

    .. code-block:: python
        conv_params = {  # They are default settings
            'convergence_energy': 1e-6,  # Eh
            'convergence_grms': 3e-4,    # Eh/Bohr
            'convergence_gmax': 4.5e-4,  # Eh/Bohr
            'convergence_drms': 1.2e-3,  # Angstrom
            'convergence_dmax': 1.8e-3,  # Angstrom
        }
        from pyscf import geometric_solver
        opt = geometric_solver.GeometryOptimizer(method)
        opt.params = conv_params
        opt.kernel()
    '''
    if isinstance(method, lib.GradScanner):
        g_scanner = method
    elif isinstance(method, GradientsBase):
        g_scanner = method.as_scanner()
    elif getattr(method, 'nuc_grad_method', None):
        g_scanner = method.nuc_grad_method().as_scanner()
    else:
        raise NotImplementedError('Nuclear gradients of %s not available' % method)
    if not include_ghost:
        g_scanner.atmlst = numpy.where(method.cell.atom_charges() != 0)[0]

    tmpf = tempfile.mktemp(dir=lib.param.TMPDIR)
    engine = PySCFEngine(g_scanner)
    engine.callback = callback
    engine.maxsteps = maxsteps
    # To avoid overwritting method.mol
    engine.cell = g_scanner.cell.copy()

    # When symmetry is enabled, the molecule may be shifted or rotated to make
    # the z-axis be the main axis. The transformation can cause inconsistency
    # between the optimization steps. The transformation is muted by setting
    # an explicit point group to the keyword mol.symmetry (see symmetry
    # detection code in Mole.build function).

    # geomeTRIC library on pypi requires to provide config file log.ini.
    if not os.path.exists(os.path.abspath(
            os.path.join(geometric.optimize.__file__, '..', 'log.ini'))):
        kwargs['logIni'] = os.path.abspath(os.path.join(__file__, '..', 'log.ini'))

    engine.assert_convergence = assert_convergence
    try:
        geometric.optimize.run_optimizer(customengine=engine, input=tmpf,
                                         constraints=constraints, **kwargs)
        conv = True
        # method.mol.set_geom_(m.xyzs[-1], unit='Angstrom')
    except NotConvergedError as e:
        lib.logger.note(method, str(e))
        conv = False
    return conv, engine.cell

def optimize(method, assert_convergence=ASSERT_CONV,
             include_ghost=INCLUDE_GHOST, constraints=None, callback=None,
             maxsteps=100, **kwargs):
    '''Optimize geometry with geomeTRIC library for the given method.

    To adjust the convergence threshold, parameters can be set in kwargs as
    below:

    .. code-block:: python
        conv_params = {  # They are default settings
            'convergence_energy': 1e-6,  # Eh
            'convergence_grms': 3e-4,    # Eh/Bohr
            'convergence_gmax': 4.5e-4,  # Eh/Bohr
            'convergence_drms': 1.2e-3,  # Angstrom
            'convergence_dmax': 1.8e-3,  # Angstrom
        }
        from pyscf import geometric_solver
        newmol = geometric_solver.optimize(method, **conv_params)
    '''
    # MRH, 07/23/2019: name all explicit kwargs for forward compatibility
    return kernel(method, assert_convergence=assert_convergence, include_ghost=include_ghost,
                  constraints=constraints, callback=callback, maxsteps=maxsteps, **kwargs)[1]

class GeometryOptimizer(lib.StreamObject):
    '''Optimize the molecular geometry for the input method.

    Note the method.mol will be changed after calling .kernel() method.
    '''
    def __init__(self, method):
        self.method = method
        self.callback = None
        self.params = {}
        self.converged = False
        self.max_cycle = 100


    @property
    def cell(self):
        return self.method.cell

    @cell.setter
    def cell(self, x):
        self.method.cell = x
        self.method.mol = x

    def kernel(self, params=None):
        if params is not None:
            self.params.update(params)
        self.converged, self.cell = \
                kernel(self.method, callback=self.callback,
                       maxsteps=self.max_cycle, **self.params)
        return self.cell
    optimize = kernel

class NotConvergedError(RuntimeError):
    pass

del (INCLUDE_GHOST, ASSERT_CONV)


if __name__ == '__main__':
    from pyscf.pbc import gto, scf, grad
    cell = gto.Cell()
    cell.atom = '''
    C 0.000000000000 0.000000000000 0.000000000000
    C 1.685068664391 1.685068664391 1.685068664391
    '''
    cell.atom= [['C', [0.0, 0.0, 0.0]], ['C', [1.685068664391,1.685068664391,1.685068664391]]]
    cell.a = '''
    0.000000000, 3.370137329, 3.370137329
    3.370137329, 0.000000000, 3.370137329
    3.370137329, 3.370137329, 0.000000000'''
    cell.basis = 'gth-szv'
    cell.verbose= 4
    cell.pseudo = 'gth-pade'
    cell.unit = 'bohr'
    cell.build()

    kpts = cell.make_kpts([1,1,3])
    mf = scf.KRHF(cell, kpts, exxdiv=None)
    mf.conv_tol = 1e-14
    mf.conv_tol_grad = 1e-8
    conv_params = {
        'convergence_energy': 1e-4,  # Eh
        'convergence_grms': 1e-5,    # Eh/Bohr
        'convergence_gmax': 1e-5,  # Eh/Bohr
        'convergence_drms': 1.2e-2,  # Angstrom
        'convergence_dmax': 1.8e-2,  # Angstrom
    }
    opt = GeometryOptimizer(mf).set(params=conv_params)#.run()
    opt.run()
    cell = opt.cell



#!/usr/bin/env python
# Copyright 2014-2019 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Restricted open-shell Hartree-Fock for periodic systems at a single k-point

See Also:
    pyscf/pbc/scf/khf.py : Hartree-Fock for periodic systems with k-point sampling
'''

import numpy as np
from pyscf import lib
from pyscf.lib import logger
from pyscf.scf import rohf as mol_rohf
from pyscf.pbc.scf import hf as pbchf
from pyscf.pbc.scf import uhf as pbcuhf
from pyscf import __config__


get_fock = mol_rohf.get_fock
get_occ = mol_rohf.get_occ
get_grad = mol_rohf.get_grad
make_rdm1 = mol_rohf.make_rdm1
energy_elec = mol_rohf.energy_elec
dip_moment = pbcuhf.dip_moment
get_rho = pbcuhf.get_rho

class ROHF(pbchf.RHF):
    '''ROHF class for PBCs.
    '''

    get_init_guess = pbcuhf.UHF.get_init_guess
    init_guess_by_chkfile = pbcuhf.UHF.init_guess_by_chkfile
    init_guess_by_minao  = mol_rohf.ROHF.init_guess_by_minao
    init_guess_by_atom   = mol_rohf.ROHF.init_guess_by_atom
    init_guess_by_huckel = mol_rohf.ROHF.init_guess_by_huckel
    init_guess_by_mod_huckel = mol_rohf.ROHF.init_guess_by_mod_huckel
    eig = mol_rohf.ROHF.eig
    get_fock = mol_rohf.ROHF.get_fock
    get_occ = mol_rohf.ROHF.get_occ
    get_grad = mol_rohf.ROHF.get_grad
    get_rho = get_rho
    make_rdm1 = mol_rohf.ROHF.make_rdm1
    energy_elec = mol_rohf.ROHF.energy_elec
    analyze = mol_rohf.ROHF.analyze
    canonicalize = mol_rohf.ROHF.canonicalize
    spin_square = mol_rohf.ROHF.spin_square
    stability = mol_rohf.ROHF.stability
    dip_moment = pbchf.SCF.dip_moment
    gen_response = pbcuhf.gen_response
    to_gpu = lib.to_gpu

    def __init__(self, cell, kpt=np.zeros(3),
                 exxdiv=getattr(__config__, 'pbc_scf_SCF_exxdiv', 'ewald')):
        pbchf.SCF.__init__(self, cell, kpt, exxdiv)
        self.nelec = None

    @property
    def nelec(self):
        if self._nelec is not None:
            return self._nelec
        else:
            cell = self.cell
            ne = cell.nelectron
            nalpha = (ne + cell.spin) // 2
            nbeta = nalpha - cell.spin
            if nalpha + nbeta != ne:
                raise RuntimeError('Electron number %d and spin %d are not consistent\n'
                                   'Note cell.spin = 2S = Nalpha - Nbeta, not 2S+1' %
                                   (ne, self.spin))
            return nalpha, nbeta
    @nelec.setter
    def nelec(self, x):
        self._nelec = x

    def dump_flags(self, verbose=None):
        pbchf.SCF.dump_flags(self, verbose)
        logger.info(self, 'number of electrons per cell  '
                    'alpha = %d beta = %d', *self.nelec)
        return self

    def get_veff(self, cell=None, dm=None, dm_last=0, vhf_last=0, hermi=1,
                 kpt=None, kpts_band=None):
        if cell is None: cell = self.cell
        if dm is None: dm = self.make_rdm1()
        if kpt is None: kpt = self.kpt
        if isinstance(dm, np.ndarray) and dm.ndim == 2:
            dm = np.asarray((dm*.5,dm*.5))
        if getattr(dm, 'mo_coeff', None) is not None:
            mo_coeff = dm.mo_coeff
            mo_occ_a = (dm.mo_occ > 0).astype(np.double)
            mo_occ_b = (dm.mo_occ ==2).astype(np.double)
            dm = lib.tag_array(dm, mo_coeff=(mo_coeff,mo_coeff),
                               mo_occ=(mo_occ_a,mo_occ_b))
        vj, vk = self.get_jk(cell, dm, hermi, kpt, kpts_band)
        vhf = vj[0] + vj[1] - vk
        return vhf

    def get_bands(self, kpts_band, cell=None, dm=None, kpt=None):
        '''Get energy bands at the given (arbitrary) 'band' k-points.

        Returns:
            mo_energy : (nmo,) ndarray or a list of (nmo,) ndarray
                Bands energies E_n(k)
            mo_coeff : (nao, nmo) ndarray or a list of (nao,nmo) ndarray
                Band orbitals psi_n(k)
        '''
        raise NotImplementedError

    def init_guess_by_1e(self, cell=None):
        if cell is None: cell = self.cell
        if cell.dimension < 3:
            logger.warn(self, 'Hcore initial guess is not recommended in '
                        'the SCF of low-dimensional systems.')
        return mol_rohf.ROHF.init_guess_by_1e(self, cell)

    def to_ks(self, xc='HF'):
        '''Convert to RKS object.
        '''
        from pyscf.pbc import dft
        return self._transfer_attrs_(dft.ROKS(self.cell, self.kpt, xc=xc))
